# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
r"""
Class that reads and prepares template libraries for use in fitting the
stellar-continuum of a spectrum.  See
:func:`available_template_libraries` for the list of default template
libraries.

The "raw" template libraries are expected to consist of 1D fits files.
The wavelength sampling can be either linear or logarithmic in
wavelength.  The reference frame of the template wavelengths must be
defined as either vacuum or air.  It is expected that the object spectra
to be fit are calibrated to vacuum wavelengths.  When preparing the
template spectra for analysis, this class will use
`pydl.goddard.astro.airtovac`_ to convert the template wavelengths to
vacuum.  Finally, one can specify that the template library is only
valid within a certain wavelength range and above a certian flux limit;
see :class:`TemplateLibraryDef`.

Preparation of the template library for use in stellar-continuum fitting
consists of a number of steps as described in the documentation of
:func:`TemplateLibrary.process_template_library`.

A template library that has been prepared for analysis is automatically
written to disk for later recovery.

Two support classes are also provided. One is a derived
:class:`mangadap.par.parset.KeywordParSet` instance that provides the
defining parameters of a DAP template library. The second is a
derived :class:`mangadap.util.bitmask.BitMask` instance that defines
the bitmasks for the template library spectra.

.. _templatelibrary-usage:

Class usage examples
--------------------

TODO: Update this for new datacube usage

Assuming you have the default directory structure setup, you can do::

    # Imports
    from mangadap.drpfits import DRPFits
    from mangadap.proc.templatelibrary import TemplateLibrary
    from matplotlib import pyplot

    # Define the DRP file
    drpf = DRPFits(7495, 12703, 'CUBE')

    # Build the template library
    tpl_lib = TemplateLibrary('M11MILES', drpf=drpf, directory_path='.')
    # Writes: ./manga-7495-12703-LOGCUBE_M11-MILES.fits

    # Plot one of the spectra
    pyplot.plot(tpl_lib.hdu['WAVE'].data, tpl_lib.hdu['FLUX'].data[0,:])
    pyplot.show()

As part of the instantiation of the :class:`TemplateLibrary` object
in the above call, the template library is prepared for use in
fitting the specified DRP file.  If the required processing has
already been done, the instantiation of the :class:`TemplateLibrary`
object simply reads the existing file.  If you do not have the
default directory structure setup, you'll need to define the paths
to, e.g., the DRP file; see :class:`mangadap.util.drpfits.DRPFits`.

If you do not want to process the template (match the spectral
resolution and sampling to that of the DRP data), you can force the
:class:`TemplateLibrary` object to only provide the "raw" spectra::

    # Imports
    from mangadap.proc.templatelibrary import TemplateLibrary
    from matplotlib import pyplot

    # Read the raw template library
    tpl_lib = TemplateLibrary('M11MILES', process=False)
    # Nothing should be written to disk

    # Plot one of the spectra
    pyplot.plot(tpl_lib.hdu['WAVE'].data[0,:], tpl_lib.hdu['FLUX'].data[0,:])
    pyplot.show()

Note that in the previous example, the wavelength data was already
one-dimensional, whereas for the raw library, the wavelength vector
can be spectrum-dependent.

In the above examples, the user has not provided a list of template
libraries, meaning that the default set available to the DAP is
used.  The default set is defined in
:func:`available_template_libraries`.  If you want to use your own
template library, you have to define its parameters using
:class:`TemplateLibraryDef`.  Currently, the template library
spectra are expected to be 1D fits files with WCS header keywords
defining the wavelength solution; see above.  Using an existing DAP
library as an example::

    # Imports
    import os
    from mangadap.proc.templatelibrary import TemplateLibraryDef
    from mangadap.proc.templatelibrary import TemplateLibrary
    from mangadap.config.defaults import dap_data_root
    from matplotlib import pyplot

    # Define the search string for the library
    search_str = os.path.join(dap_data_root(), 'stellar_templates/miles/*.fits')

    # Define the template library parameters
    tpl_par = TemplateLibraryDef(key='MILES',    # Unique keyword for the library
                            file_search=search_str, # Search string
                            fwhm=2.50,              # FWHM of resolution element (assumed const)
                            in_vacuum=False,        # Wavelength in vacuum?
                            wave_limit=numpy.array([ 3575., 7400. ]),   # Range of valid lambda
                            lower_flux_limit=0.0)   # Lower limit for valid flux

    # Read the raw template library
    tpl_lib = TemplateLibrary('MILES', tpllib_list=tpl_par, process=False)
    # Nothing should be written to disk

    # Plot one of the spectra
    pyplot.plot(tpl_lib.hdu['WAVE'].data[0,:], tpl_lib.hdu['FLUX'].data[0,:])
    pyplot.show()

Note that the keyword you use must be provided both to the parameter
set and when instantiating the :class:`TemplateLibrary` object.  In
the example above, I have not processed the library, but you can by
following a similar approach to the first example.

You can also process the spectra to a user-provided resolution and
pixel sampling, toggle on/off the renormalization of the library to
a mean flux of unity and define various paths if you're not working
within the nominal DAP directory structure.  See the optional
instantiation arguments for :class:`TemplateLibrary`.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import inspect
from pathlib import Path
import warnings
import logging

from IPython import embed

import numpy
from scipy import interpolate

from astropy.io import fits
import astropy.constants

from pydl.goddard.astro import airtovac

from ..par.parset import KeywordParSet
from ..config import defaults
from ..util.dapbitmask import DAPBitMask
from ..util.log import log_output
from ..util.fileio import read_template_spectrum, writefits_1dspec, create_symlink
from ..util.fitsutil import DAPFitsUtil
from ..util import sampling
from ..util.resolution import SpectralResolution, match_spectral_resolution
from ..util.parser import DefaultConfig
from .util import HDUList_mask_wavelengths


class TemplateLibraryDef(KeywordParSet):
    """
    Class with parameters used to define the template library.
    See :class:`mangadap.par.parset.ParSet` for attributes.

    The defined parameters are:

    .. include:: ../tables/templatelibrarydef.rst
    """
    def __init__(self, key='MILESHC', file_search='miles_cluster/*.fits', fwhm=2.50, sres_ext=None,
                 in_vacuum=False, wave_limit=None, lower_flux_limit=None, log10=False): 

        # Use the signature to get the parameters and the default values
        sig = inspect.signature(self.__class__)
        pars = list(sig.parameters.keys())
        defaults = [sig.parameters[key].default for key in pars]

        # Remaining definitions done by hand
        in_fl = [int, float]
        arr_like = [numpy.ndarray, list]
        
        values = [key, file_search, fwhm, sres_ext, in_vacuum, wave_limit, lower_flux_limit, log10]
        dtypes = [str, str, in_fl, str, bool, arr_like, in_fl, bool]
        descr = ['Keyword to distinguish the template library.',
                 'Search pattern used to find the 1D fits spectra to include in the ' \
                    'template library.  The search string must either include the full path ' \
                    'or be within the template directory of the DAP source distribution.',
                 'FWHM of the resolution element in angstroms.',
                 'Extension in the fits files with measurements of the spectral resolution as ' \
                    'a function of wavelength.  If present, this supercedes any provided FWHM.',
                 'Flag that the wavelengths of the spectra are in vacuum, not air.',
                 'Two-element array with the starting and ending wavelengths for the valid ' \
                    'spectral range of the templates.',
                 'Minimum valid flux in the template spectra.',
                 'Flag that the template spectra have been binned logarithmically in wavelength.']

        super().__init__(pars, values=values, defaults=defaults, dtypes=dtypes, descr=descr)
        self._validate()

    @classmethod
    def from_key(cls, key):
        """
        Instantiate the template parameters using a keyword only.
        """
       # Check the configuration files exist
        search_dir = defaults.dap_config_root() / 'spectral_templates'
        ini_files = sorted(list(search_dir.glob('*.ini')))
        if len(ini_files) == 0:
            raise IOError(f'Could not find any configuration files in {search_dir} !') 
        for i in range(len(ini_files)):
            cnfg = DefaultConfig(f=ini_files[i], interpolate=True)
            if key != cnfg['key']:
                continue
            wave_limit = None if cnfg['wave_limit'] is None \
                            else numpy.array(cnfg.getlist('wave_limit', evaluate=True))
            return cls(key=cnfg['key'], file_search=cnfg['file_search'],
                       fwhm=cnfg.getfloat('fwhm'), sres_ext=cnfg['sres_ext'],
                       in_vacuum=cnfg.getbool('in_vacuum'), wave_limit=wave_limit,
                       lower_flux_limit=cnfg.getfloat('lower_flux_limit'),
                       log10=cnfg.getbool('log10', default=False))

        # If the code makes it here, the keyword did not select a known template library
        raise ValueError(f'Unknown template library: {key}')

    @classmethod
    def from_dict(cls, d):
        """
        Instantiate from a dictionary.

        If the dictionary only contains the keyword for the template library,
        the function assumes the user is trying to define the library based on
        that keyword.  Otherwise, the function calls the base-class
        ``from_dict`` method.
        """
        # First make sure that the dictionary defines the library keyword
        if 'key' not in d.keys():
            raise ValueError('Template library instantiation must include a keyword identifier.')
        if list(d.keys()) == ['key']:
            return cls.from_key(d['key'])
        return super().from_dict(d)

    def _validate(self):
        """ 
        Validate the template library parameters.
        """
        if self['key'] is None:
            raise ValueError('Must define the template library keyword identifier.')
        if self['file_search'] is None:
            raise ValueError('Must define the search pattern for template library files.')
        if self['fwhm'] is None and self['sres_ext'] is None:
            raise ValueError('Template library must provide either \'fwhm\' or \'sres_ext\'.')

    def get_file_list(self):
        """
        Use the search string to find the template library fits files.

        The search string must be via the default location in the DAP
        distribution or the full path to a local library.

        The file list read by the search key is sorted for
        consistency.

        Returns:
            :obj:`list`: The sorted list of files.

        Raises:
            ValueError:
                Raised if no files are found.
        """
        # Try the DAP directory first
        root = defaults.dap_data_root() / 'spectral_templates' / self['file_search']
        if root.parent.exists():
            files = sorted(list(root.parent.glob(root.name)))
        if not root.parent.exists() or len(files) == 0:
            # Then try the provided path directly
            root = Path(self['file_search']).resolve()
            if root.parent.exists():
                files = sorted(list(root.parent.glob(root.name)))
        if not root.parent.exists() or len(files) == 0:
            raise FileNotFoundError('Unable to find template library directory or any valid files '
                                    'as given or in the DAP source distribution.')
        return files


class TemplateLibraryBitMask(DAPBitMask):
    """
    Derived class that specifies the mask bits for the template library
    data.  The maskbits defined are:

    .. include:: ../tables/templatelibrarybitmask.rst
    """
    cfg_root = 'spectral_template_bits'


class TemplateLibrary:
    r"""
    TODO: Revisit these docs.

    Object used to read, store, and prepare template libraries for
    use in analyzing object spectra.

    The default list of available libraries provided by the MaNGA DAP
    defined in :func:`available_template_libraries`. The user can
    provide their own library for use with this class provided they
    are contained in 1D fits spectra, sampled linearly in wavelength
    with the wavelength coordinates available via the WCS keywords
    (``CRPIX1``, ``CRVAL1``, ``CDELT1``), and they have an
    appropriately defined spectral resolution. See
    :class:`TemplateLibraryDef` and :func:`_build_raw_hdu`.

    The class is optimized for use in analyzing MaNGA DRP files;
    however, one can provide the necessary information so that the class
    can be used with a non-DRP spectrum.  In the latter case, the user
    must supply the velocity scale of the pixel for the logarithmically
    resampled template library, and a
    :class:`mangadap.util.resolution.SpectralResolution` object the
    defines the instrumental resolution of the spectrum/spectra to be
    analyzed.

    .. todo::

        - Docs are out of date.
        - Update the list of attributes.
        - Allow for velocity scale input instead of spectral_step?
        - Only works with DRP files that have log-linear wavelength
          binning!
        - Allow to process, where process is just to change the sampling
          or the resolution (not necessarily both).
        - Need to make this more general, removing all dependence on
          DRPFits object.  This would simplify the functionality to
          change how the resolution and sampling matching is specified.

    On initialization, if the datacube object is not provided (is
    None), the default behavior is to read the raw template library
    if read=True. If the datacube is provided, the routine will check
    for the resolution matched fits file; if it doesn't exist and
    read is ``True``, it will prepare the template library for use in
    analyzing the datacube and write the prepared library file (if
    hardcopy is True). If overwrite=True, the preparation and writing
    of the template library will be done even if the library already
    exists.

    .. todo::

        Change `sres` to be a standard input object?

    Args:
        library (:obj:`str`, :class:`TemplateLibraryDef`):
            A keyword selecting one of the template libraries distributed with
            the DAP or the definition of a user-provided library.
        cube (:class:`mangadap.datacube.datacube.DataCube`, optional):
            The datacube to be analyzed using the template library.
        match_resolution (:obj:`bool`, optional):
            Match the spectral resolution of the template library to
            the resolution provided by the ``cube``; the latter must
            be provided for this argument to have any use.
        velscale_ratio (:obj:`int`, optional):
            Resample the template spectra such that the ratio of the
            pixel scale in the provided ``cube`` is this many times
            larger than the pixels in the resampled template
            spectrum.
        sres (:class:`mangadap.util.resolution.SpectralResolution`, optional):
            The object is used simply to access the spectral
            resolution and associated wavelength coordinate vector
            needed when matching the spectral resolution of the
            template library; this is used in lieu of a datacube.
        velocity_offset (:obj:`float`, optional):
            Velocity offset to use when matching the spectral
            resolution between the template library and the galaxy
            spectra.
        min_sig_pix (:obj:`float`, optional):
            Minimum value of the LSF standard deviation in pixels
            allowed before assuming the kernel is a Delta function.
            See
            :func:`mangadap.util.resolution.match_spectral_resolution`.
        no_offset (:obj:`bool`, optional):
            Prohibit the spectral resolution matching to introduce an
            offset in spectral resolution required to match the
            relative shape of the resolution vectors. See
            :func:`mangadap.util.resolution.match_spectral_resolution`
            and
            :func:`mangadap.util.resolution.SpectralResolution.GaussianKernelDifference`.
        spectral_step (:obj:`float`, optional):
            Logarithmic (``log=True``) or linear (``log=False``) step
            in wavelength for the template library.
        log (:obj:`bool`, optional):
            Flag to force the library to be logarithmically sampled in
            wavelength.
        wavelength_range (array-like, optional):
            Force the template library to cover this spectral range.
            Unobserved spectral regions will be masked.
        renormalize (:obj:`bool`, optional):
            After processing, renormalize the flux of the full
            library to unity. The relative fluxes of the templates
            are maintained.
        output_path (:obj:`str`, `Path`_, optional):
            The path for the output file.  If None, the current working
            directory is used.
        output_file (:obj:`str`, optional):
            The name of the output file. The full path of the
            output file will be :attr:`directory_path`/:attr:`output_file`.
            If None, the default is to combine ``cube.output_root`` and the
            library key; if ``cube`` is None, it is simply the library key.
        read (:obj:`bool`, optional):
            Flag to read the template library data.
        process (:obj:`bool`, optional):
            Process (spectral resolution and sampling operations) the
            template library.  See :func:`process_template_library`.
        hardcopy (:obj:`bool`, optional):
            Flag to keep a hardcopy of the processed template library.
        symlink_dir (:obj:`str`, optional):
            Create a symlink to the file in this directory. If None,
            no symbolic link is created.
        overwrite (:obj:`bool`, optional):
            Overwrite any saved, processed template library file.
        checksum (:obj:`bool`, optional):
            Use the checksum in the fits header to confirm that the
            data has not been corrupted. The checksum is **always**
            written to the fits header when the file is created.
        loggers (:obj:`list`, optional):
            List of `logging.Logger`_ objects to log progress;
            ignored if quiet=True. Logging is done using
            :func:`mangadap.util.log.log_output`. If None, no logging
            is performed and output is just written to ``stdout``.
        quiet (:obj:`bool`, optional):
            Suppress all terminal and logging output.

    Attributes:
        bitmask (class:`mangadap.util.bitmask.BitMask`):
            Object used to toggle mask values; see
            :func:`TemplateLibraryBitMask`.
        library (:class:`TemplateLibraryDef`):
            Parameter set required to read and prepare the library.
        file_list (:obj:`list`):
            The list of files found using `glob.glob`_ and
            :attr:`file_search`.
        ntpl (:obj:`int`):
            Number of template spectra in the library
        cube (:class:`mangadap.datacube.datacube.DataCube`):
            The datacube to be analyzed using the template library.
        sres (:class:`mangadap.util.resolution.SpectralResolution`):
            The object is used simply to access the spectral resolution
            and associated wavelength coordinate vector needed when
            matching the spectral resolution of the template library;
            this is used in place of the attributes in any provided DRP
            file object.
        velocity_offset (:obj:`float`):
            Velocity offset to use when matching the spectral resolution
            between the template library and the galaxy spectra.
        spectral_step (:obj:`float`):
            Target logarithmic (``log10_sampling=True``) or linear
            (``log10_sampling=False``) step in wavelength for the
            template library.
        log10_sampling (:obj:`bool`):
            Flag that the processed template library is logarithmically
            sampled in wavelength.
        directory_path (:obj:`str`):
            The exact path to the processed template library file.
        output_file (:obj:`str`):
            The name of the file containing (to contain) the prepared
            template library output file.  The file should be found at
            :attr:`directory_path`/:attr:`output_file`.
        processed (:obj:`bool`):
            Flag that the template library has been prepared for use in
            the DAP.
        hardcopy (:obj:`bool`):
            Flag to keep a hardcopy of the processed template library.
        symlink_dir (:obj:`str`):
            Symlink created to the file in this directory
        hdu (`astropy.io.fits.HDUList`_):
            HDUList read from the DAP file

    """
    # Class attribute
    supported_libraries = ['BC03', 'BPASS', 'M11ELODIE', 'M11MARCS', 'M11MILES', 'M11STELIB',   
                           'M11STELIBZSOL', 'MASTARHC', 'MASTARHC2', 'MASTARSSP', 'MILES',
                           'MILESAVG', 'MILESHC', 'MILESTHIN', 'MIUSCAT', 'MIUSCATTHIN', 'STELIB']
    """Provides the keywords of the supported libraries."""

    def __init__(self, library, cube=None, match_resolution=True, velscale_ratio=None, sres=None,
                 velocity_offset=0.0, min_sig_pix=0.0, no_offset=True, spectral_step=None,
                 log=True, wavelength_range=None, renormalize=True, output_path=None,
                 output_file=None, read=True, process=True, hardcopy=False, symlink_dir=None,
                 overwrite=False, checksum=False, loggers=None, quiet=False):

        self.loggers = loggers
        self.quiet = quiet

        # Define the TemplateLibraryBitMask object
        self.bitmask = TemplateLibraryBitMask()

        # Define the properties needed to modify the spectral resolution
        self.sres = None
        self.velocity_offset = None
        self.min_sig_pix = None
        self.no_offset = None

        # Define the target spectral sampling properties
        self.spectral_step = None
        self.log10_sampling = None

        # Define the library
        self.library = library if isinstance(library, TemplateLibraryDef) \
                            else TemplateLibraryDef.from_key(library)
        self.file_list = None
        self.ntpl = None

        # Define the processed file and flag, and the HDUList used to
        # keep the data
        self.directory_path = None
        self.output_file = None
        self.processed = False
        self.hardcopy = hardcopy
        self.symlink_dir = None
        self.hdu = None
        self.checksum = checksum

        # Do not read the library
        if not read:
            if not self.quiet:
                log_output(self.loggers, 1, logging.INFO, 'Nothing read, by request.')
            return
       
        # Do not process the library
        if not process:
            if not self.quiet:
                log_output(self.loggers, 1, logging.INFO, 'Reading raw library without processing.')
            self._read_raw()
            return

        # Read and process the library
        self.process_template_library(cube=cube, match_resolution=match_resolution,
                                      velscale_ratio=velscale_ratio, sres=sres,
                                      velocity_offset=velocity_offset, min_sig_pix=min_sig_pix,
                                      no_offset=no_offset, spectral_step=spectral_step, log=log,
                                      wavelength_range=wavelength_range, renormalize=renormalize,
                                      output_path=output_path, output_file=output_file,
                                      hardcopy=hardcopy, symlink_dir=symlink_dir,
                                      overwrite=overwrite, loggers=loggers, quiet=quiet)

    def __getitem__(self, key):
        return self.hdu[key]

    @staticmethod
    def default_paths(library_key, cube=None, output_path=None, output_file=None):
        """
        Set the default directory and file name for the output file.

        Args:
            library_key (:obj:`str`):
                Unique library keyword.
            cube (:class:`mangadap.datacube.datacube.DataCube`, optional):
                Datacube to analyze.  The output is specific to the cube if the
                spectral resolution and spectral sampling are matched to the
                cube data.  If None, the output file name is simply the library
                key.
            output_path (:obj:`str`, `Path`_, optional):
                The path for the output file.  If None, the current working
                directory is used.
            output_file (:obj:`str`, optional):
                The name of the output file. The full path of the
                output file will be :attr:`directory_path`/:attr:`output_file`.
                If None, the default is to combine ``cube.output_root`` and the
                library key.

        Returns:
            :obj:`tuple`: Returns a `Path`_ with the output directory and a
            :obj:`str` with the output file name.
        """
        directory_path = Path('.').resolve() if output_path is None \
                                else Path(output_path).resolve()
        root = f'{library_key}' if cube is None else f'{cube.output_root}-{library_key}'
        _output_file = f'{root}.fits.gz' if output_file is None else output_file
        return directory_path, _output_file
    
    def _get_nchannels(self):
        """
        Get the maximum number of wavelength channels needed to store
        all the template spectra in a single array.

        .. todo::
            - What happens if the spectrum has an empty primary
              extension?

        Returns:
            :obj:`int`: Maximum number of pixels used by the listed
            spectra.

        Raises:
            ValueError:
                Raised if the input template spectra are not
                one-dimensional.
        """
        max_npix = 0
        for f in self.file_list:
            if fits.getval(str(f), 'NAXIS') != 1:
                raise ValueError('{0} is not one dimensional!'.format(f))
            npix = fits.getval(f, 'NAXIS1')
            if max_npix < npix:
                max_npix = npix
        return max_npix

    def _build_raw_hdu(self, npix):
        r"""
        Build the "raw" template library arrays.

        This simply reads the provided list of fits files and puts
        them into arrays of size :math:`N_{\rm tpl} \times N_{\rm
        pix}`.

        This will *force* reading of the data, even if the :attr:`hdu`
        is already initialized.

        The :attr:`hdu` will contain the appropriate extensions, but
        it is important to note that the wavelength vectors will
        **not** necessarily be the same. That is, reading of the raw
        template spectra can accommodate spectra that have different
        wavelength coordinates. Any pixels that have no data are
        masked using the 'NO_DATA' bitmask flag; see
        :func:`TemplateLibraryBitMask`.

        The spectral resolution is set using the ``fwhm`` or
        ``sres_ext`` values in the library parameters.

        .. warning::

            Currently no errors are saved because none are expected for
            the template libraries.

        Args:
            npix (:obj:`int`):
                Number of spectral channels for the output arrays
        """
        if self.hdu is not None:
            if not self.quiet:
                log_output(self.loggers, 1, logging.INFO, 'Closing existing HDUList.')
            self.hdu.close()
            self.hdu = None

        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, 'Attempting to build raw data ...')
        # Allocate the vectors
        wave = numpy.zeros((self.ntpl, npix), dtype=float)
        flux = numpy.zeros((self.ntpl, npix), dtype=float)
        sres = numpy.zeros((self.ntpl, npix), dtype=float)
        mask = numpy.zeros((self.ntpl, npix), dtype=self.bitmask.minimum_dtype())
        soff = numpy.zeros(self.ntpl, dtype=float)

        # Read and save each spectrum and mask the unobserved
        # wavelengths
        for i in range(0,self.ntpl):
            if self.library['sres_ext'] is None:
                wave_, flux_ = read_template_spectrum(str(self.file_list[i]),
                                                      log10=self.library['log10'])
                # TODO: Are the copies here needed?
                wave[i,:wave_.size] = numpy.copy(wave_)
                flux[i,:wave_.size] = numpy.copy(flux_)
                sres = wave/self.library['fwhm']
            else:
                wave_, flux_, sres_ = read_template_spectrum(str(self.file_list[i]),
                                                             sres_ext=self.library['sres_ext'],
                                                             log10=self.library['log10'])
                wave[i,:wave_.size] = numpy.copy(wave_)
                flux[i,:wave_.size] = numpy.copy(flux_)
                sres[i,:wave_.size] = numpy.copy(sres_)

            if wave_.size != npix:
                mask[i,wave_.size:] = self.bitmask.turn_on(mask[i,wave_.size:],'NO_DATA')
            if self.library['lower_flux_limit'] is not None:
                indx = numpy.invert( flux_ > self.library['lower_flux_limit'] )
                mask[i,indx] = self.bitmask.turn_on(mask[i,indx], 'FLUX_INVALID')
            if self.library['wave_limit'] is not None \
                    and self.library['wave_limit'][0] is not None:
                indx = wave[i,:].ravel() < self.library['wave_limit'][0]
                mask[i,indx] = self.bitmask.turn_on(mask[i,indx], 'WAVE_INVALID')
            if self.library['wave_limit'] is not None \
                    and self.library['wave_limit'][1] is not None:
                indx = wave[i,:].ravel() > self.library['wave_limit'][1]
                mask[i,indx] = self.bitmask.turn_on(mask[i,indx], 'WAVE_INVALID')

        # (Re)Set the HDUList object
        self._reset_hdu(wave, flux, mask, sres, soff)

        # Add some keywords to the header
        self.hdu[0].header['TPLPROC'] = (0, 'Flag that library has been processed')
        self.processed = False
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, '... done')

    def _read_raw(self):
        """
        Read the 'raw' versions of the template library.

        The "raw" library is the data before its sampling or
        resolution have been altered.

        This is a simple wrapper for
        :func:`~mangadap.proc.templatelibrary.TemplateLibraryDef._get_file_list`,
        :func:`_get_channels` and :func:`_build_raw_hdu`.
        """
        self.file_list = self.library.get_file_list()
        self.ntpl = len(self.file_list)
        npix = self._get_nchannels()
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO,
                            'Found {0} {1} templates'.format(self.ntpl, self.library['key']))
            log_output(self.loggers, 1, logging.INFO,
                            'Maximum number of wavelength channels: {0}'.format(npix))
        self._build_raw_hdu(npix)

    def _reset_hdu(self, wave, flux, mask, sres, soff):
        r"""
        (Re)Set :attr:`hdu` to a new HDUList object using the input
        arrays.  Also sets the header items indicating the version of
        the template reader and the keyword for the library.

        .. warning::

            No checking is done concerning the size of each array!

        Args:
            wave (`numpy.ndarray`_):
                Array with the wavelengths of each pixel.
            flux (`numpy.ndarray`_):
                Array with the flux in each pixel.
            mask (`numpy.ndarray`_):
                Bitmask values for each pixel.
            sres (`numpy.ndarray`_):
                Spectral resolution, :math:`R=\lambda/\delta\lambda`,
                at each pixel.
            soff (`numpy.ndarray`_):
                The spectral resolution offset for each spectrum (see
                :func:`mangadap.util.resolution.GaussianKernelDifference`).
        """
        if self.hdu is not None:
            self.hdu.close()
            self.hdu = None
        self.hdu = fits.HDUList([ fits.PrimaryHDU(),
                                  fits.ImageHDU(wave, name='WAVE'),
                                  fits.ImageHDU(flux, name='FLUX'),
                                  fits.ImageHDU(mask, name='MASK'),
                                  fits.ImageHDU(sres, name='SPECRES'),
                                  fits.ImageHDU(soff, name='SIGOFF')])
        self.hdu['PRIMARY'].header['LIBKEY'] = (self.library['key'], 'Library identifier')

    def _wavelength_range(self, flag=None):
        """
        Return the valid wavelength range for each spectrum based on the
        first and last unmasked pixel; interspersed masked regions are
        not considered.

        Args:
            flag (:obj:`str`, :obj:`list`, optional):
                Flags to consider when determining the wavelength
                range; see
                :func:`mangadap.util.bitmasks.BitMask.flagged`.

        Returns:
            `numpy.ndarray`_: Two-element vector with wavelengths of
            the first and last valid pixels.
        """
        indx = numpy.where(numpy.invert(self.bitmask.flagged(self.hdu['MASK'].data, flag=flag)))
        return numpy.array([numpy.amin(self.hdu['WAVE'].data[indx]),
                            numpy.amax(self.hdu['WAVE'].data[indx])]).astype(float)

    def _minimum_sampling(self):
        """
        Return the minimum sampling of the available wavelength vectors.
        
        Returns:
            :obj:`float`: Minimum sampling of all (will just be one
            if the library has been processed) wavelength vectors.
        """
        if self.processed:
            return self.spectral_step

        nspec = self.hdu['WAVE'].data.shape[0]
        minimum_step = sampling.spectral_coordinate_step(self.hdu['WAVE'].data[0,:].ravel(),
                                                         log=self.library['log10'])
        for i in range(1,nspec):
            step = sampling.spectral_coordinate_step(self.hdu['WAVE'].data[0,:].ravel(),
                                                     log=self.library['log10'])
            if minimum_step > step:
                minimum_step = step
        return minimum_step

    def _rebin_masked(self, i, flag, fullRange, rmsk_lim=0.5):
        """
        Determine the mask value to adopt for a rebinned spectrum by
        rebinning the mask pixels, setting a masked pixel to unity,
        and an unmasked pixel to zero. After rebinning, any pixel
        with a value of larger than ``rmsk_lim`` is masked; otherwise
        it is left unmasked.

        Although the function can be used with multiple flags, its
        intended use is to determine which pixels should be masked
        with a specific flag.

        Args:
            i (:obj:`int`):
                Index of the spectrum to be rebinned.
            flag (:obj:`str`, :obj:`list`):
                Flags to consider when determining which pixels to
                mask; see
                :func:`mangadap.util.bitmasks.BitMask.flagged`.
            fullRange (`numpy.ndarray`_):
                Two-element array with the wavelength range for the
                rebinned spectrum.
            rmsk_lim (:obj:`float`):
                Limit of the rebinned mask value that is allowed
                before considering the pixel as masked.

        Returns:
            `numpy.ndarray`_: Boolean array indicating which pixels
            in the rebinned spectrum should be masked.
        """
        unity = numpy.ma.MaskedArray(numpy.ones(self.hdu['WAVE'].data.shape[1], dtype=float),
                                     mask=self.bitmask.flagged(self.hdu['MASK'].data[i,:],
                                                               flag=flag))
        r = sampling.Resample(unity, x=self.hdu['WAVE'].data[i,:], inLog=self.library['log10'],
                              newRange=fullRange, newLog=self.log10_sampling,
                              newdx=self.spectral_step)
        return r.outf < rmsk_lim

    # TODO: Having this return the redshift is a bit weird.
    def _modify_spectral_resolution(self):
        """
        Modify the spectral resolution to match the provided
        :attr:`sres`.

        Returns:
            :obj:`float`:
                Redshift used when matching the spectral resolution.
        """
        # Calculate to use as an offset of the match to the spectral
        # resolution. Used to better match the spectral resolution to
        # at the *observed* wavelengths of the object spectrum to which
        # the TemplateLibrary will be fit.
        redshift = self.velocity_offset/astropy.constants.c.to('km/s').value

        # Mask wavelengths where the spectral resolution will have to be
        # extrapolated.
        sres_wave = self.sres.wave()
        wavelim = numpy.array([sres_wave[0]/(1.+redshift), sres_wave[-1]/(1.+redshift)])
        self.hdu = HDUList_mask_wavelengths(self.hdu, self.bitmask, 'SPECRES_EXTRAP', wavelim,
                                            invert=True)

        # Match the resolution of the templates.  ivar is returned, but
        # is always None because ivar is not provided to
        # match_spectral_resolution
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, 'Modifying spectral resolution ... ')

#        flux_inp = self.hdu['FLUX'].data.copy()

        self.hdu['FLUX'].data, self.hdu['SPECRES'].data, self.hdu['SIGOFF'].data, res_mask, \
                ivar = match_spectral_resolution(self.hdu['WAVE'].data, self.hdu['FLUX'].data,
                                                 self.hdu['SPECRES'].data, sres_wave/(1.+redshift),
                                                 self.sres.sres(), min_sig_pix=self.min_sig_pix,
                                                 no_offset=self.no_offset,
                                                 log10=self.library['log10'], new_log10=True,
                                                 quiet=self.quiet)
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, '... done')

        # Mask any pixels where the template resolution was too low to
        # match to the galaxy resolution
        self.hdu['MASK'].data[res_mask == 1] = \
            self.bitmask.turn_on(self.hdu['MASK'].data[res_mask == 1], 'SPECRES_LOW')

        # Mask any pixels where the template flux is 0 or below
        indx = numpy.invert(self.hdu['FLUX'].data > 0)
        if numpy.any(indx):
            self.hdu['MASK'].data[indx] = self.bitmask.turn_on(self.hdu['MASK'].data[indx],
                                                               'SPECRES_NOFLUX')

        # Function returns the redshift used for the spectral
        # resolution matching. TODO: This should be input instead!
        return redshift


    def _process_library(self, wavelength_range=None, renormalize=True):
        """
        Process the template library for use in analyzing object
        spectra.   See :func:`process_template_library`.

        .. todo::

            - Add wavelength coordinate WCS information to the
              appropriate extension headers.
            - Include 'step' as an argument
            - Include limit on covering fraction used to mask data as an
              argument

        Args:
            wavelength_range (array-like, optional):
                A two-element object with the lower and upper limits
                for the wavelength range of the library.
            renormalize (:obj:`bool`, optional):
                Renormalize the flux of the full library to unity.
                Relative fluxes of spectra in the library are
                maintained.
        """
        # Convert to vacuum wavelengths
        # TODO: This conversion to vacuum wavelengths is non-linear.
        # This is accounted for in the resampling of the spectra (the
        # pixel size can be non-uniform), but *not* the resolution
        # matching.  This is okay for DR15 because we never resolution
        # match the data, but it will matter when we start to do that by
        # switching templates between the stellar and gas kinematics
        # fit.
        if not self.library['in_vacuum']:
            self.hdu['WAVE'].data = airtovac(self.hdu['WAVE'].data)

        #---------------------------------------------------------------
        # Modify the spectral resolution to a target function, if one
        # has been provided.
        redshift = 0.0 if self.sres is None else self._modify_spectral_resolution()

        #---------------------------------------------------------------
        # Resample the templates to a logarithmic binning step.
        #
        # Typically this will be run to match the sampling of the
        # template library to that of a set of galaxy spectra to be fit.
        # However, the sampling can be left alone and this part will
        # force all the spectra to have the same wavelength range.
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, 'Matching sampling ... ')

        # Number of angstroms per pixel
        ang_per_pix = numpy.array([sampling.angstroms_per_pixel(w, log=self.library['log10'],
                                                                base=10.,
                                                                regular=self.library['in_vacuum'])
                                        for w in self.hdu['WAVE'].data])

        # TODO: I think this kludge was necessary because of a bug in
        # angstroms_per_pixel.  If regular was True and log was False,
        # it only returns a single integer, otherwise it returned a
        # vector.  I've changed angstroms_per_pixel, but have left the
        # kludge below for now for testing.

        #KHRR KLUDGE
        if len(ang_per_pix.shape) == 1:
            warnings.warn('Executing KLUDGE!  Bug fix insufficient.')
            tmp_ang_per_pix = numpy.empty(self.hdu['SPECRES'].data.shape)
            for ind in range(len(ang_per_pix)):
                tmp_ang_per_pix[ind,:] = ang_per_pix[ind]
            ang_per_pix = tmp_ang_per_pix

        # Number of pixels per resolution element
        pix_per_fwhm = numpy.ma.divide(self.hdu['WAVE'].data,
                                       self.hdu['SPECRES'].data * ang_per_pix)

        # TODO: Put in a check of whether or not it needs to do the
        # resampling!

        # The raw spectra are allowed to have wavelength ranges that
        # differ.  First, determine the wavelength range that encloses
        # all spectra.  Only ignore pixels that were flagged as having
        # no data.
        no_data_flags = ['NO_DATA', 'SPECRES_NOFLUX']
        fullRange = self._wavelength_range(flag=no_data_flags) if wavelength_range is None \
                            else numpy.array(wavelength_range).astype(float)

        # Get the spectral step if it hasn't been set yet
        if self.spectral_step is None:
            self.spectral_step = self._minimum_sampling()

        # Get the number of pixels needed
        npix, _fullRange = sampling.grid_npix(rng=fullRange, dx=self.spectral_step,
                                              log=self.log10_sampling)

        # Any pixels without data after resampling are given a value
        # that is the minimum flux - 100 so that they can be easily
        # identified afterward.  The minimum flux is:
        min_flux = numpy.amin(self.hdu['FLUX'].data.ravel())
        # the observed pixels are
        no_data = self.bitmask.flagged(self.hdu['MASK'].data, flag='NO_DATA')

        # Now resample the spectra.  First allocate the arrays
        flux = numpy.zeros((self.ntpl, npix), dtype=float)
        sres = numpy.zeros((self.ntpl, npix), dtype=float)
        mask = numpy.zeros((self.ntpl, npix), dtype=self.bitmask.minimum_dtype())

        for i in range(self.ntpl):
            # Rebin the observed wavelength range
            r = sampling.Resample(self.hdu['FLUX'].data[i,:], mask=no_data[i,:],
                                  x=self.hdu['WAVE'].data[i,:], inLog=self.library['log10'],
                                  newRange=fullRange, newLog=self.log10_sampling,
                                  newdx=self.spectral_step, step=False)
            
            # Save the result
            wave = r.outx
            flux[i,:] = r.outy
            indx = r.outf < 0.9
            # TODO: Set the flux to 0?
            flux[i,indx] = 0.0
            mask[i,indx] = self.bitmask.turn_on(mask[i,indx], 'NO_DATA')

            # Resample the spectral resolution by simple interpolation.
            sres[i,:] = interpolate.interp1d(self.hdu['WAVE'].data[i,:],
                                             self.hdu['SPECRES'].data[i,:], assume_sorted=True,
                                             fill_value='extrapolate')(wave)
            sres[i,indx] = 0.0

            # Recalculate the number of pixels per fwhm
            # - Number of angstroms per pixel
            _ang_per_pix = sampling.angstroms_per_pixel(wave, log=self.log10_sampling, base=10.,
                                                        regular=True)
            # - Number of pixels per resolution element
            _pix_per_fwhm = numpy.ma.divide(wave, sres[i,:] * _ang_per_pix)

            indx = _pix_per_fwhm < 2
            if numpy.sum(indx) > 0:
                warnings.warn('Resampling results in resolution below the two pixel limit!')
                sres[i,indx] = wave[indx]/2./_ang_per_pix[indx]
                mask[i,indx] = self.bitmask.turn_on(mask[i,indx], 'SPECRES_2PIXEL')
            #-----------------------------------------------------------

            # Rebin the masks, bit-by-bit:
            # Pixels outside the wavelength limits
            indx = self._rebin_masked(i, 'WAVE_INVALID', fullRange, rmsk_lim=0.1)
            mask[i,indx] = self.bitmask.turn_on(mask[i,indx], 'WAVE_INVALID')
            # Pixels below the flux limit
            indx = self._rebin_masked(i, 'FLUX_INVALID', fullRange, rmsk_lim=0.1)
            mask[i,indx] = self.bitmask.turn_on(mask[i,indx], 'FLUX_INVALID')
            # Pixels that required an extrapolation of the spectral
            # resolution
            indx = self._rebin_masked(i, 'SPECRES_EXTRAP', fullRange, rmsk_lim=0.1)
            mask[i,indx] = self.bitmask.turn_on(mask[i,indx], 'SPECRES_EXTRAP')
            # Pixels that had a spectral resolution that was too low to
            # match the galaxy resolution
            indx = self._rebin_masked(i, 'SPECRES_LOW', fullRange, rmsk_lim=0.1)
            mask[i,indx] = self.bitmask.turn_on(mask[i,indx], 'SPECRES_LOW')
            # Pixels that no flux because of the wavelength limits set
            # by the resolution matching.
            indx = self._rebin_masked(i, 'SPECRES_NOFLUX', fullRange, rmsk_lim=0.1)
            mask[i,indx] = self.bitmask.turn_on(mask[i,indx], 'SPECRES_NOFLUX')
            
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, '... done')
            log_output(self.loggers, 1, logging.INFO,
                       'After resampling (target): {0}'.format(self.spectral_step))
            log_output(self.loggers, 1, logging.INFO,
                       'After resampling (calculated): {0}'.format(
                            sampling.spectral_coordinate_step(wave, log=True)))

        # Finally, we need to (again) trim the wavelength range to
        # exclude pixels with invalid or empty flux.
        edges = numpy.ma.notmasked_edges(numpy.ma.MaskedArray(numpy.ones(mask.shape),
                            mask=self.bitmask.flagged(mask,
                            flag=['NO_DATA', 'WAVE_INVALID', 'FLUX_INVALID', 'SPECRES_NOFLUX'])),
                                         axis=1)
        s = numpy.amax(edges[0][1])
        e = numpy.amin(edges[1][1])+1
        wave = wave[s:e]
        flux = flux[:,s:e]
        mask = mask[:,s:e]
        sres = sres[:,s:e]

        # Normalize the templates to the mean flux value after excluding
        # any flagged pixels.
        if renormalize:
            indx = numpy.invert(self.bitmask.flagged(mask, flag=['NO_DATA', 'WAVE_INVALID', 
                                                                 'FLUX_INVALID']))
            if numpy.sum(indx) == 0:
                if not self.quiet:
                    warnings.warn('All pixels masked.  Unable to renormalize TemplateLibrary.')
                flux_norm = 1.0
            else:
                flux_norm = numpy.mean(flux[numpy.where(indx)])
            flux /= flux_norm
        else:
            flux_norm = 1.0

        # Reset the HDUList object
        self._reset_hdu(wave, flux, mask, sres, self.hdu['SIGOFF'].data)

        # Update the header with the redshift offset, the flux
        # normalization, and flag the data as having been prepared for
        # fitting the DRP data
        self.hdu[0].header['ZGUESS'] = (redshift,'Redshift used when matching spectral resolution')
        self.hdu[0].header['FLXNRM'] = (flux_norm,'Flux normalization')
        self.hdu[0].header['TPLPROC'] = 1
        self.processed = True

    def file_name(self):
        """Return the name of the processed file."""
        return self.output_file

    def file_path(self):
        """Return the full path to the processed file."""
        if self.directory_path is None or self.output_file is None:
            return None
        return self.directory_path / self.output_file

    def read_raw_template_library(self, library=None):
        """
        Read the identified template library.  If all the arguments are
        the default, the preset attributes from the initialization of
        the object are used.

        Args:
            library (:obj:`str`, :class:`TemplateLibraryDef`, optional):
                Keyword selecting the library to use or the full set of defining
                parameters.
        """
        if library is not None:
            self.library = library if isinstance(library, TemplateLibraryDef) \
                                else TemplateLibraryDef.from_key(library)
        self._read_raw()

    def process_template_library(self, library=None, cube=None, match_resolution=True,
                                 velscale_ratio=None, sres=None, velocity_offset=0.0,
                                 min_sig_pix=0.0, no_offset=True, spectral_step=None, log=True,
                                 wavelength_range=None, renormalize=True, output_path=None,
                                 output_file=None, hardcopy=False, symlink_dir=None,
                                 overwrite=False, loggers=None, quiet=False):
        r"""
        Process the template library for use in analyzing spectra.

        .. todo::
            - Docs need to be updated
            - type checking
            - If a DRP file is provided, the processing to a logarithmic
              binning is done by default (log=True).  But linearly
              sampled DRP data are available, so need to have
              :class:`mangadap.util.drpfits.DRPFits` return the spectral
              sampling type.
            - Documentation needs updating!
        
        Primary steps are to:

            - Read the raw 1D fits files; see :func:`_read_raw`.

            - Convert the wavelengths to vacuum, if necessary; see
              `pydl.goddard.astro.airtovac`_.

            - Mask wavelengths outside the rest wavelength range of the
              datacube spectra, due to need to extrapolate these values; see
              :func:`mangadap.util.bitmasks.HDUList_mask_wavelengths`.

            - Match the spectral resolution of the template to that of
              the datacube (if ``match_resolution`` is True); see
              :func:`mangadap.util.resolution.match_spectral_resolution`.

            - Mask the template pixels where the spectral resolution was
              too low to match to datacube; see
              :func:`mangadap.util.bitmasks.BitMask.turn_on`.

            - Force a common wavelength range and sampling for all
              templates, where the sampling is forced to match the
              sampling of the DRP spectra; see
              :func:`mangadap.util.sampling.grid_pix`. The masks are
              appropriately resampled as well; see
              :func:`_rebin_masked`.

        .. warning::

            The routine **does not** check that that an existing
            processed file or the existing object has been processed
            using the same datacube, velocity_offset, velscale, or
            sres input. If unsure, use overwrite=True.
        
        Args:
            library (:obj:`str`, :class:`TemplateLibraryDef`, optional):
                Keyword selecting the library to use or the full set of defining
                parameters.
            cube (:class:`mangadap.datacube.datacube.DataCube`, optional):
                The datacube to be analyzed using the template library.
            match_resolution (:obj:`bool`, optional):
                Match the spectral resolution of the template library to
                the resolution provided by the ``cube``; the latter must
                be provided for this argument to have any use.
            velscale_ratio (:obj:`int`, optional):
                Resample the template spectra such that the ratio of the
                pixel scale in the provided ``cube`` is this many times
                larger than the pixels in the resampled template
                spectrum.
            sres (:class:`mangadap.util.resolution.SpectralResolution`, optional):
                The object is used simply to access the spectral
                resolution and associated wavelength coordinate vector
                needed when matching the spectral resolution of the
                template library; this is used in lieu of a datacube.
            velocity_offset (:obj:`float`, optional):
                Velocity offset to use when matching the spectral
                resolution between the template library and the galaxy
                spectra.
            min_sig_pix (:obj:`float`, optional):
                Minimum value of the LSF standard deviation in pixels
                allowed before assuming the kernel is a Delta function.
                See
                :func:`mangadap.util.resolution.match_spectral_resolution`.
            no_offset (:obj:`bool`, optional):
                Prohibit the spectral resolution matching to introduce an
                offset in spectral resolution required to match the
                relative shape of the resolution vectors. See
                :func:`mangadap.util.resolution.match_spectral_resolution`
                and
                :func:`mangadap.util.resolution.SpectralResolution.GaussianKernelDifference`.
            spectral_step (:obj:`float`, optional):
                Logarithmic (``log=True``) or linear (``log=False``) step
                in wavelength for the template library.
            log (:obj:`bool`, optional):
                Flag to force the library to be logarithmically sampled in
                wavelength.
            wavelength_range (array-like, optional):
                Force the template library to cover this spectral range.
                Unobserved spectral regions will be masked.
            renormalize (:obj:`bool`, optional):
                After processing, renormalize the flux of the full
                library to unity. The relative fluxes of the templates
                are maintained.
            output_path (:obj:`str`, `Path`_, optional):
                The path for the output file.  If None, the current working
                directory is used.
            output_file (:obj:`str`, optional):
                The name of the output file. The full path of the
                output file will be :attr:`directory_path`/:attr:`output_file`.
                If None, the default is to combine ``cube.output_root`` and the
                library key; if ``cube`` is None, it is simply the library key.
            hardcopy (:obj:`bool`, optional):
                Flag to keep a hardcopy of the processed template library.
            symlink_dir (:obj:`str`, optional):
                Create a symlink to the file in this directory. If None,
                no symbolic link is created.
            overwrite (:obj:`bool`, optional):
                Overwrite any saved, processed template library file.
            loggers (:obj:`list`, optional):
                List of `logging.Logger`_ objects to log progress;
                ignored if quiet=True. Logging is done using
                :func:`mangadap.util.log.log_output`. If None, no logging
                is performed and output is just written to ``stdout``.
            quiet (:obj:`bool`, optional):
                Suppress all terminal and logging output.

        Raises:
            ValueError:
                If the sampling is undefined, or if a hardcopy is
                requested but the output file could not be defined.
        """
        # Initialize the reporting
        if loggers is not None:
            self.loggers = loggers
        self.quiet = quiet

        # TODO: Not sure we should allow the library key to change in this
        # way...
        if library is not None:
            self.library = library if isinstance(library, TemplateLibraryDef) \
                                else TemplateLibraryDef.from_key(library)

        # Ignore velscale_ratio if it is set to unity
        if velscale_ratio is not None and velscale_ratio == 1:
            velscale_ratio = None

        if cube is not None:
            # Use the provided datacube to set the sampling
            self.sres = None
            self.log10_sampling = cube.log
            self.spectral_step = sampling.spectral_coordinate_step(cube.wave, log=cube.log)

            # Use the datacube to set the spectral resolution, if
            # resolution matching was requested. The fiducial spectral
            # resolution is set to be the median spectral resolution of
            # all the spectra that are not fully masked.
            if match_resolution:
                # Copy the spectral resolution to a masked array
                self.sres = cube.copy_to_masked_array(attr='sres')
                # Only use the spectra that are not fully masked
                indx = numpy.sum(numpy.invert(numpy.ma.getmaskarray(self.sres)), axis=1) > 0
                # Get the median resolution
                self.sres = numpy.median(self.sres.data[indx], axis=0)
                # Instantiate the resolution object
                self.sres = SpectralResolution(cube.wave, self.sres, log10=self.log10_sampling)
        else:
            # Use the provided input values
            self.sres = sres
            self.log10_sampling = log
            self.spectral_step = spectral_step

        self.velocity_offset = velocity_offset
        self.min_sig_pix = min_sig_pix
        self.no_offset = no_offset

        # Adjust for the velocity scale ratio between the template and
        # object data to be fit
        if not self.log10_sampling and velscale_ratio is not None:
            raise ValueError('velscale_ratio only valid with logarithmically sampled spectra.')
        if velscale_ratio is not None:
            self.spectral_step /= velscale_ratio

        # (Re)Set the output paths
        self.directory_path, self.output_file \
                = TemplateLibrary.default_paths(self.library['key'], cube=cube,
                                                output_path=output_path, output_file=output_file)
        self.hardcopy = hardcopy
        self.symlink_dir = symlink_dir

        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, '-'*50)
            log_output(self.loggers, 1, logging.INFO,
                       f'Template library output path: {self.directory_path}')
            log_output(self.loggers, 1, logging.INFO,
                       f'Template library output file: {self.output_file}')

        # Check that the path for or to the file is defined
        ofile = self.file_path()
        if self.hardcopy and ofile is None:
            raise ValueError('File path for output file is undefined!')

        # Read and use a pre-existing file
        if self.hardcopy and ofile.exists() and not overwrite:
            if not self.quiet:
                log_output(self.loggers, 1, logging.INFO, 'Reading existing file')
            self.hdu = DAPFitsUtil.read(ofile, checksum=self.checksum)
            self.file_list = self.library.get_file_list()
            self.ntpl = self.hdu['FLUX'].data.shape[0]
            self.processed = True
            # Make sure the symlink exists
            if self.symlink_dir is not None:
                create_symlink(str(ofile), self.symlink_dir, overwrite=overwrite)
            if not self.quiet:
                log_output(self.loggers, 1, logging.INFO, '-'*50)
            return

        # Warn the user that the file will be overwritten
        if self.hardcopy and ofile.exists():
            if not self.quiet:
                warnings.warn(f'Overwriting existing file: {ofile}')

        # Read the raw data
        self._read_raw()

        # Process the library
        self._process_library(wavelength_range=wavelength_range, renormalize=renormalize)

        # Write the fits file
        if self.hardcopy:
            DAPFitsUtil.write(self.hdu, str(ofile), overwrite=overwrite, checksum=True,
                              symlink_dir=self.symlink_dir, loggers=self.loggers, quiet=self.quiet)
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, '-'*50)

    def single_spec_to_fits(self, i, ofile, overwrite=True):
        """
        Write one of the template spectra to a 1D fits file.

        Output is a one-dimensional fits file with a single extension
        containing the flux measurements and header keywords needed
        to determine the wavelength of each pixel.

        Args:
            i (:obj:`int`):
                Index of the spectrum in the avaiable list to output.
            ofile (:obj:`str`):
                Name of the file to write
            overwrite (:obj:`bool`, optional):
                Overwrite any existing file.

        Raises:
            ValueError:
                Raised if the selected index is not available.
        """
        if i < 0 or i >= self.ntpl:
            raise ValueError('Invalid index ({0})!  Library contains {1} spectra.'.format(i,
                                                                                        self.ntpl))
        wave = self.hdu['WAVE'].data if self.processed else self.hdu['WAVE'].data[i,:]
        log = self.log10_sampling if self.processed else self.library['log10']
        writefits_1dspec(ofile, numpy.log10(wave[0]) if log else wave[0], self.spectral_step,
                         self.hdu['FLUX'].data[i,:], overwrite=overwrite)





# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""

Stack some spectra!

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""

from IPython import embed

import numpy
from scipy import sparse, interpolate
from matplotlib import pyplot, rc

#from astropy.io import fits
import astropy.constants

from ..par.parset import KeywordParSet
from ..util.covariance import Covariance
from ..util.filter import interpolate_masked_vector
from ..util.sampling import Resample, spectral_coordinate_step


class SpectralStackPar(KeywordParSet):
    r"""
    Class with parameters used to set how to stack a set of spectra.
    See :class:`mangadap.par.parset.ParSet` for attributes.

    .. todo::
        Allow for sigma rejection.

    The defined parameters are:

    .. include:: ../tables/spectralstackpar.rst

    .. warning::

        Velocity registration is currently *not* implemented.

    """
    def __init__(self, operation='mean', register=False, cz=None, covar_mode='none',
                 covar_par=None):
        in_fl = [ int, float ]
        ar_like = [ numpy.ndarray, list ]
        op_options = SpectralStack.operation_options()
        covar_options = SpectralStack.covariance_mode_options()
        
        pars = ['operation', 'register', 'cz',  'covar_mode',   'covar_par']
        values = [operation, register, cz, covar_mode, covar_par]
        defaults = ['mean', False, None, 'none', None]
        options = [op_options, None, None, covar_options, None]
        dtypes = [str, bool, ar_like, str, in_fl+ar_like]
        descr = ['Operation to perform for the stacked spectrum.  See ' \
                    ':func:`SpectralStack.operation_options` for the available operation options.',
                 'Flag to register the spectra by deshifting them based on their observed *cz* ' \
                    'velocities.  This is done before adding them based on a provided ' \
                    'prior measurement of the velocities.',
                 'List of measured *cz* velocities used to register the spectra.',
                 'Describes how to incorporate covariance into the spectral stacking.  ' \
                    'See :func:`SpectralStack.covariance_mode_options` for the available options.',
                 'The parameter(s) needed to perform a given method of handling the ' \
                    'covariance.  See :func:`SpectralStack.covariance_mode_options` for the ' \
                    'available options.']
        super().__init__(pars, values=values, defaults=defaults, options=options, dtypes=dtypes,
                         descr=descr)

    def toheader(self, hdr):
        """
        Copy the information to a header.

        Args:
            hdr (`astropy.io.fits.Header`_):
                Header object to write to.

        Returns:
            `astropy.io.fits.Header`_ : Edited header object.
        """
        hdr['STCKOP'] = (self['operation'], 'Stacking operation')
        hdr['STCKVREG'] = (str(self['register']), 'Spectra shifted in velocity before stacked')
        hdr['STCKCRMD'] = (str(self['covar_mode']), 'Stacking treatment of covariance')
        hdr['STCKCRPR'] = (str(self['covar_par']), 'Covariance parameter(s)')
        return hdr

    def fromheader(self, hdr):
        """
        Copy the information from the header

        Args:
            hdr (`astropy.io.fits.Header`_):
                Header object to read from.
        """
        self['operation'] = hdr['STCKOP']
        self['register'] = eval(hdr['STCKVREG'])
        self['covar_mode'] = hdr['STCKCRMD']
        self['covar_par'] = eval(hdr['STCKCRPR'])


class SpectralStack:
    r"""
    Class whose primary function is to stack a set of spectra while
    treating covariance between spectra.

    See :func:`covariance_mode_options` for available methods of
    accounting for covariance.

    Class also approximates the resulting spectral resolution of the
    stacked data.

    .. todo::
        List attributes

    """
    def __init__(self):
        # Keep the matrix used to bin the spectra. This will be a
        # scipy.sparse.csr_matrix.
        self.rebin_T = None

        # Internal arrays for callback
        self.wave = None
        self.flux = None
        self.fluxsqr = None
        self.fluxmean = None
        self.fluxsdev = None
        self.npix = None
        self.ivar = None
        self.sres = None
        self.covar = None

    @staticmethod
    def _check_covariance_type(covariance_mode, covar, ivar):
        """
        Check that the covariance variable has the correct type for
        the given mode.

        Args:
            covariance_mode (:obj:`str`):
                Covariance handling mode; see
                :func:`covariance_mode_options`.
            covar (None, :obj:`float`, :class:`mangadap.util.covariance.Covariance`):
                The object to check the type against the covariance
                handling mode.
            ivar (None, object):
                Inverse variance. Only check performed is whether or
                not this is None.
            
        Returns:
            :obj:`bool`: Flag that type is correct.
        """
        if covariance_mode == 'none':
            return True
        if covariance_mode in ['calibrate', 'channels', 'wavelengths'] and ivar is None:
            return False
        if covariance_mode == 'calibrate' and not isinstance(covar, float):
            return False
        if covariance_mode == 'calibrate':
            return True
        if not isinstance(covar, Covariance):
            return False
        return covar.dim == 3

    @staticmethod
    def _check_covariance_shape(covariance_mode, covar, nwave, nspec):
        """
        Check that the input covariance object has the correct shape
        for the given mode.

        Args:
            covariance_mode (:obj:`str`):
                Covariance handling mode; see
                :func:`covariance_mode_options`.
            covar (None, :obj:`float`, :class:`mangadap.util.covariance.Covariance`):
                The object to check the type against the covariance
                handling mode.
            nwave (:obj:`int`):
                Number of wavelength channels.
            nspec (:obj:`int`):
                Number of spectra.

        Returns:
            :obj:`bool`: Flag that the covariance data has the
            correct shape.
        """
        if covariance_mode in [ 'none', 'calibrate']:
            return True
        if covariance_mode in [ 'full', 'approx_correlation' ] \
                and (covar.dim != 3 or covar.shape[-1] != nwave):
            return False
        if covar.shape[0] != nspec:
            return False
        return True

    @staticmethod
    def _get_input_mask(flux, ivar=None, mask=None, dtype=bool):
        """
        Construct the baseline mask using the input arrays.

        Args:
            flux (`numpy.ndarray`_, `numpy.ma.MaskedArray`_):
                Input flux array.
            ivar (`numpy.ndarray`_):
                Inverse variance array. Mask excludes any inverse
                variance values that are not greater than 0.
            mask (`numpy.ndarray`_):
                Baseline boolean mask array.
            dtype (data type):
                Output data type for array.
        
        Returns:
            `numpy.ndarray`_: Boolean mask array converted to
            ``dtype``.
        """
        inp_mask = numpy.zeros(flux.shape, dtype=bool) if mask is None else mask
        if isinstance(flux, numpy.ma.MaskedArray):
            inp_mask |= numpy.ma.getmaskarray(flux)
        if ivar is not None:
            inp_mask |= numpy.invert(ivar>0)
        return inp_mask.astype(dtype)

    @staticmethod
    def _check_input_sres(sres, nspec):
        r"""
        Check the shape and type of the input spectral resolution.

        If the input is a masked array, interpolate over the mask.

        Args:
            sres (`numpy.ndarray`_, `numpy.ma.MaskedArray`_):
                Input spectral resolution data. If None, None is
                returned.
            nspec (:obj:`int`):
                Expected number of spectra.

        Returns:
            `numpy.ndarray`_: The interpolated spectral resolution
            vectors. If the input ``sres`` is a single vector, the
            spectral resolution is repeated for each spectrum. Shape
            is :math:`(N_{\rm spec}, N_{\rm wave})`. If the input
            ``sres`` is None, instead None is returned.
        """
        if sres is None:
            return None
        if sres.ndim > 2:
            raise ValueError('Spectral resolution must be 2D or less.')
        if sres.ndim == 2 and nspec > 0 and sres.shape[0] != nspec:
            raise ValueError('Spectral resolution array is not the correct shape.')
        _sres = sres
        if isinstance(sres, numpy.ma.MaskedArray):
            _sres = numpy.apply_along_axis(interpolate_masked_vector, 1,
                                           sres.reshape(1,-1) if sres.ndim == 1 else sres)
            if sres.ndim == 1:
                _sres = _sres[0,:]
        if _sres.ndim == 2 or nspec == 0:
            return _sres
        return numpy.array([_sres]*nspec)

    def _set_rebin_transfer_matrix(self, binid, binwgt=None):
        r"""
        Construct the transfer matrix that rebins the spectra.

        The shape of the transfer matrix is :math:`(N_{\rm bin}
        \times N_{\rm spec})` with the expectation that the spectrum
        flux array has shape :math:`(N_{\rm spec} \times N_{\rm
        wave})`.

        The binned spectra are calculated by matrix multiplication,
        :math:`\mathbf{B} = \mathbf{T} \times \mathbf{F}` such that
        the covariance matrix can be calculated as :math:`\mathbf{C}
        = \mathbf{T} \times \mathbf{\Sigma} \times \mathbf{T}^{\rm
        T}`, where :math:`\mathbf{\Sigma}` is the covariance matrix
        in the flux array, :math:`\mathbf{F}`.

        If weighting, the sum of the weights is normalized to the
        number of points included in the bin.

        Args:
            binid (`numpy.ndarray`_):
                Index, one per spectrum in the flux array, for the
                binned spectrum. Indices of less than one are
                ignored.
            binwgt (`numpy.ndarray`_, optional):
                List of weights for the spectra. If not provided, the
                weights are uniform.
        """
        nspec = binid.size
        valid = binid > -1
        unique_bins = numpy.unique(binid[valid])
        nbin = len(unique_bins)
        self.rebin_T = numpy.zeros((nbin,nspec), dtype=float)
        for j in range(nbin):
            indx = binid == unique_bins[j]
            self.rebin_T[j,indx] = 1.0 if binwgt is None else \
                                    binwgt[indx]*numpy.sum(indx)/numpy.sum(binwgt[indx])
        self.rebin_T = sparse.csr_matrix(self.rebin_T)

    def _stack_without_covariance(self, flux, ivar=None, sres=None, linear_sres=False):
        """
        Stack the spectra, ignoring any covariance that may or may not
        exist.

        This method calculates the sum of the flux, flux^2, and
        determines the number of pixels in the sum. Sets
        :attr:`flux`, :attr:`fluxsqr`, :attr:`npix`, :attr:`ivar`,
        :attr:`sres`. The stored data is always based on the **sum**
        of the spectra in the stack.
        
        Args:
            flux (`numpy.ma.MaskedArray`_):
                Flux array.
            ivar (:obj:`numpy.ma.MaskedArray`, optional):
                The inverse variance array.
            sres (`numpy.ma.MaskedArray`_, optional):
                1D or 2D spectral resolution as a function of
                wavelength for all or each input spectrum. Default is
                to ignore any spectral resolution data.
            linear_sres (:obj:`bool`, optional):
                Construct the combined resolution as the mean of the
                input instrumental FWHM (1/sres). If False, the
                quadratic mean is used (i.e., rms =
                sqrt(mean(square))); this mode is formally correct
                for the second moment of the combined LSF (however,
                note that this not the same as the accuracy of a
                Gaussian fit to the combined line profile; cf. Law et
                al. 2020).
        """
        # Convert the stack transform matrix to a dense array
        rt = self.rebin_T.toarray()
        # Relevant shapes
        nstack, nspec = rt.shape
        nwave = flux.shape[-1]

        # Get the number of pixels included in the stack of each spectrum.
        gpm = numpy.logical_not(numpy.ma.getmaskarray(flux)).astype(int)
        self.npix = numpy.dot(rt, gpm)

        # The gymnastics below are done primarily to ensure that stacks
        # that only include a single spectrum do *not* lose their
        # masked values in the array with the stacked spectra. I.e.,
        # these spectra are simply copied to the stacked array,
        # whereas numpy.dot is used for the stack calculations. NOTE:
        # numpy.ma.dot was abandoned because it is *much* slower than
        # the simple multiplication by the gpm done below.

        # Calculate the number of spectra in each stack
        nspec_per_stack = numpy.sum((rt > 0).astype(int), axis=1)

        # Construct the output vectors
        self.flux = numpy.ma.zeros((nstack, nwave), dtype=float)
        self.fluxsqr = numpy.ma.zeros((nstack, nwave), dtype=float)
        self.ivar = None
        if ivar is not None:
            self.ivar = numpy.ma.zeros((nstack, nwave), dtype=float)
        self.sres = None
        if sres is not None:
            Tc = numpy.sum(rt, axis=1)
            Tc[numpy.logical_not(Tc>0)] = 1.0
            self.sres = numpy.ma.zeros((nstack, nwave), dtype=float)
        
        # Find the stacked spectra that include more than one input
        # spectrum
        stacki, speci = numpy.where((nspec_per_stack[:,None] > 1) & (rt > 0))
        # Do the stacking
        self.flux[stacki] = numpy.dot(rt[stacki,:][:,speci], gpm[speci]*flux.data[speci])
        self.fluxsqr[stacki] = numpy.dot(rt[stacki,:][:,speci],
                                         numpy.square(gpm[speci]*flux.data[speci]))
        if self.ivar is not None:
            self.ivar[stacki] = numpy.ma.power(numpy.dot(numpy.square(rt[stacki,:][:,speci]),
                                        numpy.ma.divide(gpm[speci], ivar[speci]).filled(0.0)), -1.)
        if self.sres is not None:
            self.sres[stacki] = numpy.ma.power(numpy.dot(rt[stacki,:][:,speci],
                                                    numpy.ma.power(sres[speci], -1).filled(0.0))
                                                / Tc[stacki,None], -1) if linear_sres \
                                    else numpy.ma.power(numpy.dot(rt[stacki,:][:,speci],
                                                    numpy.ma.power(sres[speci], -2).filled(0.0))
                                                / Tc[stacki,None], -0.5)

        # Copy over the data from the "stacks" that only include single
        # spectra
        stacki, speci = numpy.where((nspec_per_stack[:,None] == 1) & (rt > 0))
        self.flux[stacki] = flux[speci]
        self.fluxsqr[stacki] = numpy.square(flux[speci])
        if self.ivar is not None:
            self.ivar[stacki] = ivar[speci]
        if self.sres is not None:
            self.sres[stacki] = sres[speci]

        # Ensure that pixels with zero contributions are masked;
        # spectral resolution vectors are *not* masked.
        self.flux[self.npix == 0] = numpy.ma.masked
        self.fluxsqr[self.npix == 0] = numpy.ma.masked
        if self.ivar is not None:
            self.ivar[self.npix == 0] = numpy.ma.masked

    def _stack_with_covariance(self, flux, covariance_mode, covar, ivar=None, sres=None):
        """
        Stack the spectra and incorporate covariance.
        
        Args:
            flux (`numpy.ma.MaskedArray`_):
                Flux array.
            covariance_mode (:obj:`str`):
                Covariance handling mode; see
                :func:`covariance_mode_options`.
            covar (None, :obj:`float`, :class:`mangadap.util.covariance.Covariance`):
                The relevant covariance object that must match the
                needs of the covariance handling mode.
            ivar (:obj:`numpy.ma.MaskedArray`, optional):
                The inverse variance array. Must not be None if
                ``covariance_mode`` is 'channels' or 'wavelengths'.
            sres (`numpy.ma.MaskedArray`_, optional):
                1D or 2D spectral resolution as a function of
                wavelength for all or each input spectrum. Default is
                to ignore any spectral resolution data.
        """
        # First stack without covariance. This sets self.flux,
        # self.fluxsqr, self.npix, self.ivar, self.sres.
        self._stack_without_covariance(flux, ivar=ivar, sres=sres)

        # If calibrating the noise based on the equation,
        #   Noise_corrected = Noise_nocovar * (1 + f * log10(N))
        # apply it and return
        if covariance_mode == 'calibrate':
            if self.ivar is not None:
                self.ivar /= numpy.square(1.0 + covar*numpy.ma.log10(self.npix))
            return

        # Check that the code knows what to do otherwise
        # TODO: Isn't this check done elsewhere?
        if covariance_mode not in ['channels', 'wavelengths', 'approx_correlation', 'full']:
            raise ValueError('Unknown covariance mode: {0}'.format(covariance_mode))

        # Recalibrate the error based on a selection of covariance
        # channels
        recalibrate_ivar = covariance_mode in ['channels', 'wavelengths']
        if recalibrate_ivar and self.ivar is None:
            raise ValueError('Must provide ivar to recalibrate based on covar.')

        # Setup for output
        nbin = self.flux.shape[0]
        nchan = covar.shape[-1]
        self.covar = numpy.empty(nchan, dtype=sparse.csr_matrix)
        variance_ratio = numpy.ma.zeros( (nbin,nchan), dtype=float) \
                            if recalibrate_ivar else None

        # Calculate the covariance in the stack
        for i in range(nchan):
            j = covar.input_indx[i]
            cov_array = covar.full(channel=j)
            self.covar[i] = sparse.triu(self.rebin_T.dot(cov_array.dot(self.rebin_T.T))).tocsr()

            # Get the variance ratio
            if recalibrate_ivar:
                variance_ratio[:,i] = self.covar[i].diagonal() * self.ivar[:,j]
                variance_ratio[numpy.ma.getmaskarray(self.ivar)[:,j],i] = numpy.ma.masked

        self.covar = Covariance(self.covar, input_indx=covar.input_indx)

        # Set ivar by recalibrating the existing data
        if recalibrate_ivar:
            ratio = numpy.ma.median(variance_ratio, axis=1 )
            # TODO: Add a debug mode for this?
#            self._recalibrate_ivar_figure(ratio, ofile='ivar_calibration.pdf')
            self.ivar = numpy.ma.power(ratio, -1.0)[:,None] * self.ivar
            return

        # Get the inverse variance from the full covariance matrix
        self.ivar = numpy.ma.power(self.covar.variance(), -1.) #.filled(0.0)


    def _recalibrate_ivar_figure(self, ratio, ofile=None):
        """
        Make a plot showing the computed inverse variance
        recalibration based on the covariance matrices.

        Args:
            ratio (`numpy.ndarray`_):
                Ratio between the error with and without covariance.
            ofile (:obj:`str`, optional):
                File for the plot.  If None, output to the screen.
        """
        font = { 'size' : 20 }
        rc('font', **font)

        w,h = pyplot.figaspect(1)
        fig = pyplot.figure(figsize=(1.5*w,1.5*h))

        ax = fig.add_axes([0.2,0.3,0.7,0.4])
        ax.minorticks_on()
        ax.tick_params(which='major', length=10, direction='in', top=True, right=True)
        ax.tick_params(which='minor', length=5, direction='in', top=True, right=True)
        ax.grid(True, which='major', color='0.8', zorder=0, linestyle='-', lw=0.5)

        nbin = numpy.sum(self.rebin_T.toarray(), axis=1)
        mod_nbin = numpy.arange(0, numpy.amax(nbin), 0.1)+1
        ax.scatter(nbin, numpy.sqrt(ratio), marker='.', s=50, lw=0, color='k', zorder=3)
        ax.plot(mod_nbin, 1+1.62*numpy.log10(mod_nbin), color='C3', zorder=2)

        ax.text(0.5, -0.18, r'$N_{\rm bin}$', horizontalalignment='center',
                verticalalignment='center', transform=ax.transAxes)
        ax.text(-0.14, 0.5, r'$f_{\rm covar}$',
                horizontalalignment='center', verticalalignment='center',
                transform=ax.transAxes, rotation='vertical')
        ax.text(0.96, 0.1, '8249-12705', horizontalalignment='right',
                verticalalignment='center', transform=ax.transAxes)

        if ofile is None:
            pyplot.show()
        else:
            fig.canvas.print_figure(ofile, bbox_inches='tight')
        fig.clear()
        pyplot.close(fig)

    def _covar_in_mean(self):
        """
        Compute the covariance in the mean spectrum by propagating the
        division by the number of pixels through the covariance matrix.

        Returns:
            :class:`~mangadap.util.covariance.Covariance`: Covariance
            in the mean spectrum. Returns None if :attr:`covar` is
            None.
        """
        if self.covar is None:
            return None
        nchan = self.covar.shape[-1]
        nbin = self.flux.shape[0]
        inpix = numpy.ma.power(self.npix, -1.)
        covar = numpy.empty(nchan, dtype=sparse.csr_matrix)
        for i in range(nchan):
            j = self.covar.input_indx[i]
            _inpix = inpix[:,j,None]*inpix[None,:,j]
            covar[i] = sparse.triu(self.covar.toarray(channel=j) * _inpix).tocsr()
        return Covariance(covar, input_indx=self.covar.input_indx)
            
    def _get_stack_mean(self):
        """
        Convert the summed stack to the mean stacked spectra using
        the internal data.

        Returns:
            :obj:`tuple`: See the return statement for :func:`stack`.
        """
        _ivar = None if self.ivar is None else self.ivar * numpy.square(self.npix)
        return self.wave, self.fluxmean, self.fluxsdev, self.npix, _ivar, self.sres, \
                    self._covar_in_mean()

    @staticmethod
    def operation_options():
        """
        Return the allowed stacking operations.
        
        Current operations are:
            - ``mean``: Construct the mean of the spectra
            - ``sum``: Construct the spectrum sum.

        Returns:
            :obj:`list`: List of available operations.
        """
        return ['mean', 'sum']

    @staticmethod
    def covariance_mode_options(par_needed=False):
        r"""
        Return the list of allowed covariance options.

        The two parameters ``covar_mode`` and ``covar_par`` (see
        :class:`SpectralStackPar`) set how covariance is accounted
        for in the stacking procedure. The valid options are:

            - ``none``: The noise in the stacked spectrum is a
              nominal propagation of the error assuming no
              covariance. No parameters needed.

            - ``calibrate``: Where :math:`N_{\rm bin}` is the number
              of binned spaxels and :math:`\alpha` as a provided
              parameter, the spectral noise is calibrated following:

              .. math::

                    n_{\rm calib} = n_{\rm nominal} (1 + \alpha \log\
                    N_{\rm bin})

            - ``channels``: The noise vector of each stacked spectrum
              is adjusted based on the mean ratio of the nominal and
              formally correct calculations of the noise measurements
              over a number of spectral channels. The channels are
              drawn from across the full spectral range. The number
              of channels to use is a defined parameter. The
              covariance matrix must be provided to :func:`stack`.

            - ``wavelengths``: Functionally equivalent to
              ``channels``; however, the channels to use are set by a
              list of provided wavelengths. The covariance matrix
              must be provided to :func:`stack`.

            - ``approx_correlation``: Approximate the covariance
              matrix using a Gaussian description of the correlation
              between pixels. See
              :func:`mangadap.datacube.datacube.DataCube.approximate_correlation_matrix`.
              The value of :math:`\sigma` provides for the Gaussian
              desciption of :math:`\rho_{ij}` in the correlation
              matrix. The covariance matrix must be provided to
              :func:`stack`.

            - ``full``: The full covariance cube is calculated and
              the noise vectors are constructed using the formally
              correct calculation. No parameters needed. The
              covariance matrix must be provided to :func:`stack`.

        Returns:
            :obj:`list`: List of the allowed modes.
        """
        modes = ['calibrate', 'approx_correlation', 'channels', 'wavelengths']
        if par_needed:
            return modes
        return modes + ['none', 'full']

    @staticmethod
    def parse_covariance_parameters(mode, par):
        """
        Parse the parameters needed for the treatment of the
        covariance when stacking spectra.
    
        Args:
            mode (:obj:`str`):
                Mode to use. Must be an allowed mode; see
                :func:`covariance_mode_options`.
            par (:obj:`str`):
                String representation of the parameters for the
                specified mode.

        Returns:
            :obj:`float`, :obj:`list`: Parameters parsed from the
            input string for the designated covariance mode.

        Raises:
            TypeError:
                Raised if the input parameter could not be converted
                to a float as needed by the specified mode.
            ValueError:
                Raised if the mode is not recognized.
        """
        mode_options = SpectralStack.covariance_mode_options()
        if mode not in mode_options:
            raise ValueError('Mode not among valid options: {0}.\nOptions are: {1}'.format(mode,
                                                                                    mode_options))
        if mode in ['none', 'full']:
            return None
        if mode in ['calibrate', 'approx_correlation']:
            try:
                return float(par)
            except:
                raise TypeError('Could not convert to float: {0}'.format(par))
        if mode == 'channels':
            return int(par) #[ int(e.strip()) for e in par.split(',') ]
        if mode == 'wavelengths':
            return [ float(e.strip()) for e in par.split(',') ]

    @staticmethod
    def min_max_wave(wave, cz):
        r"""
        Determine the minimum and maximum of all shifted wavelength
        ranges.

        Args:
            wave (array-like):
                Original wavelengths.  Should be 1D.
            cz (:obj:`float`, array-like):
                The redshift of one or more spectra to be removed. Each
                element is applied to the wavelength vector to determine the
                maximum wavelength range required to full sample all
                deshifted spectra.

        Returns:
            :obj:`tuple`: Two floats with the minimum and maximum
            wavelengths.

        Raises:
            ValueError:
                Raised if either ``wave`` or ``cz`` have the wrong
                dimensionality.
        """
        _wave = numpy.atleast_1d(wave)
        if _wave.ndim != 1:
            raise ValueError('Wavelength vector should be 1D!')
        _cz = numpy.atleast_1d(cz)
        if _cz.ndim != 1:
            raise ValueError('Velocity should be a float or 1D vector.')
        _wave = numpy.array([numpy.amin(_wave), numpy.amax(_wave)])
        _shifted_wave = _wave[None,:]/(1.+_cz[:,None]/astropy.constants.c.to('km/s').value)
        return numpy.amin(_shifted_wave), numpy.amax(_shifted_wave)

    @staticmethod
    def register(wave, cz, flux, ivar=None, mask=None, sres=None, log=False, base=10.0,
                 keep_range=False, flim=0.5):
        r"""
        Register a set of spectra to the same wavelength range given a
        set of measured velocities.

        .. todo::
            - Allow for correction for deredshifting flux.

        Args:
            wave (`numpy.ndarray`_):
                Single wavelength vector for all input spectra. Must
                be 1D with shape :math:`(N_{\rm wave},)`.
            cz (:obj:`float`, array-like):
                The measured :math:`cz` velocities of all spectra or each
                spectrum individually. The "registration" deshifts all the
                spectra such that :math:`cz = 0`.
            flux (`numpy.ndarray`_, `numpy.ma.MaskedArray`_):
                Flux array to register. Must be 2D with shape
                :math:`(N_{\rm spec},N_{\rm wave})`.
            ivar (`numpy.ndarray`_, `numpy.ma.MaskedArray`_, optional):
                Inverse variance in the spectrum fluxes. Shape must
                match ``flux``.
            mask (`numpy.ndarray`_, optional):
                Boolean array with values in ``flux`` to mask.
                Default assumes nothing is masked. If ``flux`` and/or
                ``ivar`` are masked array, this is included in a
                union with those masks.  Shape must match ``flux``.
            sres (`numpy.ndarray`_, `numpy.ma.MaskedArray`_, optional):
                1D or 2D spectral resolution as a function of
                wavelength for all or each input spectrum. Default is
                to ignore any spectral resolution data. If a masked
                array, the masked pixels are interpolated.
            log (:obj:`bool`, optional):
                Flag that the wavelength vector is sampled
                geometrically in wavelength.
            base (:obj:`float`, optional):
                If the wavelength vector is geometrically sampled,
                this is the base of the logarithmic step.
            keep_range (:obj:`bool`, optional):
                When registering the wavelengths of the shifted
                spectra, keep the identical spectral range as input.
            flim (:obj:`float`, optional):
                Mask any pixels in the output flux arrays that are
                covered by less than this fraction by the input
                masked flux array.

        Returns:
            :obj:`tuple`:Returns four arrays: (1) the new wavelength
            vector common to all spectra; (2) the new masked flux
            array; (3) the new masked inverse variance array, which
            will be None if no inverse variances are provided to the
            function; and (4) the new spectral resolution vectors,
            which will be None if no spectral resolution vectors are
            provided.

        Raises:
            ValueError:
                Raised if the wavelength or velocity offset vectors are
                not one-dimensional, if the flux array is not
                two-dimensional, if the inverse variance or mask arrays
                do not have the same shape as the flux array, or if the
                number of wavelengths does not match the second axis of
                the flux array.
        """
        # Check the input
        if len(wave.shape) != 1:
            raise ValueError('Input wavelength array must be one-dimensional.')
        if len(flux.shape) != 2:
            raise ValueError('Input flux array must be two-dimensional.  To register a single ' \
                             'flux vector, use mangadap.util.sampling.Resample.')
        if flux.shape[1] != wave.size:
            raise ValueError('Flux array shape does not match the wavelength data.')
        if ivar is not None and ivar.shape != flux.shape:
            raise ValueError('Input inverse variance array must have the same shape as flux array.')
        if mask is not None and mask.shape != flux.shape:
            raise ValueError('Input mask array must have the same shape as flux array.')
        if sres is not None and sres.shape != wave.shape and sres.shape != flux.shape:
            raise ValueError('Input spectral resolution data has incorrect shape.')

        # Get the mask
        inp_mask = SpectralStack._get_input_mask(flux, ivar=ivar, mask=mask)

        # Get the spectral resolution (always a masked array)
        inp_sres = SpectralStack._check_input_sres(sres, flux.shape[0])

        # Output spectral range. If keep_range is True, this is the same as the
        # input range; otherwise, this selects the maximum range necessary to
        # accommodate all velocity shifts.
        outRange = [wave[0], wave[-1]] if keep_range \
                        else list(SpectralStack.min_max_wave(wave, cz))

        # Sampling (logarithmic or linear)
        dw = spectral_coordinate_step(wave, log=log, base=base)

        # Calculate the 1-sigma error
        ferr = None if ivar is None else numpy.ma.power(ivar, -0.5)

        # Resample the flux and error
        _cz = numpy.atleast_1d(cz)
        if len(_cz) != 1 and len(_cz) != flux.shape[0]:
            raise ValueError('Must provide one velocity for all spectra or one velocity per '
                             'spectrum.')
        if len(_cz) == 1:
            # Offsetting all spectra by a single velocity
            _wave = wave/(1+cz/astropy.constants.c.to('km/s').value)
            resamp = Resample(flux, e=ferr, mask=inp_mask, x=_wave, inLog=log, newRange=outRange,
                              newdx=dw, base=base)
            _flux = resamp.outy
            _flux[resamp.outf < flim] = numpy.ma.masked
            if ivar is None:
                _ivar = None
            else:
                _ivar = numpy.ma.power(resamp.oute, -2)
                _ivar[numpy.ma.getmaskarray(_flux)] = numpy.ma.masked

            # Interpolate the spectral resolution at the location of the new
            # wavelength vector
            _sres = None if inp_sres is None \
                        else interpolate.interp1d(_wave, inp_sres, axis=1, assume_sorted=True,
                                                  fill_value='extrapolate')(resamp.outx)
            return resamp.outx, _flux, _ivar, _sres

        # Need to resample each vector independently

        # Get the shifted wavelengths
        _wave = wave[None,:]/(1.+_cz[:,None]/astropy.constants.c.to('km/s').value)

        # Resample the first spectrum. Done outside of a loop to make sure to
        # get the length of the new spectra.
        resamp = Resample(flux[0], e=None if ferr is None else ferr[0], mask=inp_mask[0],
                          x=_wave[0], inLog=log, newRange=outRange, newdx=dw, base=base)

        # Get the new wavelength array (should be the same for all spectra)
        newwave = resamp.outx

        # Init the flux array and apply the mask
        _flux = numpy.ma.MaskedArray(numpy.zeros((flux.shape[0],resamp.outy.size), dtype=float))
        _flux[0] = resamp.outy
        _flux[0,resamp.outf < flim] = numpy.ma.masked

        # Init the error array
        _ferr = numpy.zeros_like(_flux, dtype=float)
        if ferr is not None:
            _ferr[0] = resamp.oute

        # Init the spectral resolution vector(s)
        _sres = None
        if inp_sres is not None:
            _sres = numpy.zeros_like(_flux.data, dtype=float)
            _sres[0] = interpolate.interp1d(wave, inp_sres[0], assume_sorted=True,
                                            fill_value='extrapolate')(resamp.outx)

        # Loop over the remaining spectra
        for i in range(1,flux.shape[0]):
            resamp = Resample(flux[i], e=None if ferr is None else ferr[i], mask=inp_mask[i],
                              x=_wave[i], inLog=log, newx=newwave)
            _flux[i] = resamp.outy
            _flux[i,resamp.outf < flim] = numpy.ma.masked
            if ferr is not None:
                _ferr[i] = resamp.oute
            if inp_sres is not None:
                _sres[i] = interpolate.interp1d(_wave[i], inp_sres[i], assume_sorted=True,
                                                fill_value='extrapolate')(newwave)

        # Compute the inverse variance
        if ivar is None:
            _ivar = None
        else:
            _ivar = numpy.ma.power(_ferr, -2)
            _ivar[numpy.ma.getmaskarray(_flux)] = numpy.ma.masked

        # Done
        return newwave, _flux, _ivar, _sres

    @staticmethod
    def build_covariance_data(cube, covariance_mode, covariance_par):
        """
        Construct the covariance data needed by the specified
        covariance mode.

        Args:
            cube (:class:`mangadap.datacube.datacube.DataCube`):
                Datacube object to bin.
            covariance_mode (:obj:`str`):
                Mode to use. Must be an allowed mode; see
                :func:`covariance_mode_options`.
            covariance_par (:obj:`int`, :obj:`float`, :obj:`list`):
                Parameters required by the specified mode.

        Returns:
            None, :obj:`float`,
            :class:`mangadap.util.covariance.Covariance`: Covariance
            data relevant to the selected mode.

        Raises:
            ValueError:
                Raised if the covariance mode is not valid or if a
                covariance matrix cannot be computed for the provided
                datacube.
            TypeError:
                Raised if the covariance parameters do not have the
                correct type.
        """
        # Check that the covariance data is correct
        if covariance_mode is not None \
                and covariance_mode not in SpectralStack.covariance_mode_options():
            raise ValueError('Unrecognized method for covariance: {0}'.format(covariance_mode))
        if covariance_mode is None:
            covariance_mode = 'none'
        if covariance_mode == 'none':
            return None
        
        if covariance_mode == 'calibrate':
            return covariance_par

        # TODO: I don't think this catches every use case...
        if not cube.can_compute_covariance:
            raise ValueError('Cannot compute covariance for the provided datacube.')

        if covariance_mode in [ 'channels', 'wavelengths' ]:
            if covariance_mode == 'channels':
                if not isinstance(covariance_par, int):
                    raise TypeError('Unable to handle \'channels\' covariance parameter with ' \
                                    'type: {0}'.format(type(covariance_par)))
                _covariance_par = numpy.linspace(0,cube.nwave-1,num=covariance_par).astype(int)
            else:
                _covariance_par = covariance_par
                if isinstance(_covariance_par, float):
                    _covariance_par = numpy.array([covariance_par])
                if isinstance(_covariance_par, list):
                    _covariance_par = numpy.array(covariance_par)
                if not isinstance(_covariance_par, numpy.ndarray):
                    raise TypeError('Unable to handle covariance parameter of type: {0}'.format(
                                    type(covariance_par)))

                # Convert the wavelengths to channel numbers
                _covariance_par = numpy.unique(numpy.argsort(numpy.absolute(
                                                cube.wave[:,None]-_covariance_par[None,:]),
                                                             axis=1)[0,:])
            return cube.covariance_cube(channels=_covariance_par)

        if covariance_mode == 'approx_correlation':
            if not isinstance(covariance_par, float):
                raise TypeError('For approximate correlation matrix, provide sigma as a float.')
            return cube.covariance_cube(sigma_rho=covariance_par)

        if covariance_mode == 'full':
            warnings.warn('Sit back.  This is going to take a while...')
            return cube.covariance_cube()

        raise ValueError('Unrecognized covariance method: {0}'.format(covariance_mode))


    def stack_datacube(self, cube, binid, par=None):
        r"""
        Wrapper function for :func:`stack` that accepts a datacube
        object.

        .. todo::
            - List the required elements of ``par``.

        Args:
            cube (:class:`mangadap.datacube.datacube.DataCube`):
                Datacube to stack.
            binid (:obj:`numpy.ndarray`):
                Indices of the bin in which to place each spectrum.
                Shape must be the flattened spatial shape of the
                datacube; i.e., :math:`(N_{\rm spec},)`.
            par (:class:`SpectralStackPar`, optional):
                Set of parameters used to define how to stack the
                spectra. See :func:`stack`. Does not need to be
                provided on initialization, but a parameter set is
                required for all the stacking routines.

        Returns:
            :obj:`tuple`: See the return statement for :func:`stack`.
        """
        flux = cube.copy_to_masked_array(flag=cube.do_not_stack_flags())
        ivar = cube.copy_to_masked_array(attr='ivar', flag=cube.do_not_stack_flags())
        sres = cube.copy_to_array(attr='sres')
        covar = SpectralStack.build_covariance_data(cube, par['covar_mode'], par['covar_par'])

        # Make sure all the inverse variance values are valid
        indx = numpy.invert(ivar > 0)
        if numpy.any(indx):
            flux.mask |= indx
            ivar.mask |= indx

        return self.stack(cube.wave, flux, binid=binid, ivar=ivar, log=True, keep_range=True) \
                    if par is None else \
                    self.stack(cube.wave, flux, operation=par['operation'], binid=binid, ivar=ivar,
                               sres=sres, cz=par['cz'], log=True,
                               covariance_mode=par['covar_mode'], covar=covar, keep_range=True)

    def stack(self, wave, flux, operation='mean', binid=None, binwgt=None, ivar=None, mask=None,
              sres=None, cz=None, log=False, base=10.0, covariance_mode=None, covar=None,
              keep_range=False):
        r"""
        Stack a set of spectra.

        The method also sets all the returned variables to the
        internal attributes; however, they are always kept as the
        sum, not the mean, of the spectra, regardless of the value of
        ``operation``.

        .. todo:
            - Allow for renormalization of spectra before stacking.
              Where and how, TBD.
            - If only one spectrum returned, return flux array as a
              single vector?

        Args:
            wave (`numpy.ndarray`_):
                Single wavelength vector for all input spectra.
            flux (`numpy.ndarray`_, `numpy.ma.MaskedArray`_):
                Spectrum flux values. Shape must be :math:`(N_{\rm
                spec}, N_{\rm wave})`.
            operation (:obj:`str`, optional):
                Stacking operation to perform. See
                :func:`operation_options`.
            binid (`numpy.ndarray`_, optional):
                Indices of the bin in which to place each spectrum.
                Shape is :math:`(N_{\rm spec},)`. If not provided,
                all the spectra will be combined.
            binwgt (`numpy.ndarray`_, optional):
                Weights for each of the spectra. All weights must be
                positive. A binwgt of 0 is considered identical to
                masking the spectrum. Shape is :math:`(N_{\rm
                spec},)`. If None, weights are uniform.
            ivar (`numpy.ndarray`_, `numpy.ma.MaskedArray`_, optional):
                Inverse variance in the spectrum fluxes. Shape must
                match ``flux``.
            mask (`numpy.ndarray`_, optional):
                Binary mask values for the spectrum fluxes
                (False=unmasked; True=masked). Shape must match
                ``flux``. Default assumes no pixel mask.
            sres (`numpy.ndarray`_, `numpy.ma.MaskedArray`_, optional):
                1D or 2D spectral resolution as a function of
                wavelength for all or each input spectrum. Shape must
                be :math:`(N_{\rm wave},)` or :math:`(N_{\rm
                spec},N_{\rm wave})`. If provided as a masked array,
                masked pixels are replaced with the interpolated
                resolution at that location.
            cz (`numpy.ndarray`_, optional):
                Vector with measured :math:`cz` velocities used to deshift
                the spectra before stacking. See :func:`register`.
            log (:obj:`bool`, optional):
                Flag that the wavelength vector is geometrically stepped
                in wavelength.
            base (:obj:`float`, optional):
                If the wavelength vector is geometrically stepped, this
                is the base of the logarithmic step.
            covariance_mode (:obj:`str`, optional):
                Keyword for method to use for dealing with covariance;
                see :func:`covariance_mode_options`.
            covar (:obj:`float`, :class:`mangadap.util.covariance.Covariance`, optional):
                Covariance object to use, which must match the
                expectation from the covariance mode.  See
                :func:`covariance_mode_options` and
                :func:`_check_covariance_type`.
            keep_range (:obj:`float`, optional):
                When registering the wavelengths of the shifted spectra,
                keep the identical spectral range as input.

        Returns:
            :obj:`tuple`: Returns seven objects: the wavelength
            vector, the stacked flux, the standard deviation in the
            stacked flux, the number of spectra in each stacked
            pixel, the stacked inverse variance, the stacked spectral
            resolution, and the stacked covariance.
            
        Raises:
            ValueError:
                Raised if the wavelength vector is not
                one-dimensional, if the flux array is not
                two-dimensional, if the ancillary arrays (inverse
                variance, weights, mask) do not have the same shape
                as the flux array, if any of the weights are
                negative, or if the number of wavelengths does not
                match the second axis of the flux array. Also raised
                if the covariance mode is not recognized; see
                :func:`covariance_mode_options`.
            TypeError:
                Raised if the object type of ``covar`` does not match
                what is expected given ``covariance_mode``.
            NotImplementedError:
                Raised if the requesting to both velocity register
                the spectra and correct the error vectors for spatial
                covariance.
        """
        # Check the input shapes
        if len(wave.shape) != 1:
            raise ValueError('Input wavelength vector must be one-dimensional!')
        nwave = wave.size
        if len(flux.shape) != 2:
            raise ValueError('Can only stack two-dimensional matrices.')
        if flux.shape[1] != nwave:
            raise ValueError('Flux array shape does not match the wavelength data.')
        if ivar is not None and flux.shape != ivar.shape:
            raise ValueError('Shape of the inverse-variance array must match the flux array.')
        if mask is not None and flux.shape != mask.shape:
            raise ValueError('Shape of the mask array must match the flux array.')
        if sres is not None and sres.shape != wave.shape and sres.shape != flux.shape:
            raise ValueError('Input spectral resolution data has incorrect shape.')

        nspec = flux.shape[0]
        if binid is not None and binid.size != nspec:
            raise ValueError('Length of binid must match the number of input spectra.')
        if binwgt is not None:
            if binwgt.size != nspec:
                raise ValueError('Length of binwgt must match the number of input spectra.')
            if numpy.any(binwgt < 0):
                raise ValueError('Weights cannot be negative.')

        # Check that the covariance data is correct
        if covariance_mode is not None \
                and covariance_mode not in SpectralStack.covariance_mode_options():
            raise ValueError('Unrecognized method for covariance: {0}'.format(covariance_mode))
        if covariance_mode is None:
            covariance_mode = 'none'
        if not SpectralStack._check_covariance_type(covariance_mode, covar, ivar):
            raise TypeError('Incorrect covariance and/or inverse variance object type for input ' \
                            'mode:\n mode: {0}\n input covar type: {1}\n input ivar' \
                            ' type: {2}'.format(covariance_mode, type(covar), type(ivar)))
        if not SpectralStack._check_covariance_shape(covariance_mode, covar, nwave, nspec):
            raise ValueError('Covariance object has incorrect shape for use with specified mode.')
        if isinstance(covar, Covariance) and cz is not None:
            raise NotImplementedError('Currently cannot both velocity register and apply ' \
                                      'covariance matrix calculation!')
        
        # Get the masked, velocity registered flux and inverse variance
        # arrays
        if cz is None:
            _mask = SpectralStack._get_input_mask(flux, ivar=ivar, mask=mask)
            _flux = numpy.ma.MaskedArray(flux, mask=_mask)
            _ivar = None if ivar is None else numpy.ma.MaskedArray(ivar, mask=_mask)
            _sres = SpectralStack._check_input_sres(sres, flux.shape[0])
            self.wave = wave
        else:
            self.wave, _flux, _ivar, _sres = self.register(wave, cz, flux, ivar=ivar, mask=mask,
                                                           sres=sres, log=log, base=base,
                                                           keep_range=keep_range)

        # Calculate the transfer matrix
        self._set_rebin_transfer_matrix(numpy.zeros(nspec, dtype=int) 
                                            if binid is None else binid, binwgt=binwgt)

        # Stack the spectra with or without covariance
        if covariance_mode == 'none':
            self._stack_without_covariance(_flux, ivar=_ivar, sres=_sres)
        else:
            self._stack_with_covariance(_flux, covariance_mode, covar, ivar=_ivar, sres=_sres)

        # Calculate the standard deviation in the flux, even if the flux
        # operation is to sum the data
        self.fluxmean = self.flux/self.npix
        self.fluxsdev = numpy.ma.sqrt((self.fluxsqr/self.npix - numpy.square(self.fluxmean))
                                            * self.npix*numpy.ma.power((self.npix-1), -1.))

        # Interpolate across any masked pixels in the spectral
        # resolution array, if provided and requested.
        if self.sres is not None:
            if numpy.sum(self.sres.mask) == 0:
                self.sres = self.sres.data
            else:
                outshape = self.sres.shape
                self.sres = numpy.apply_along_axis(interpolate_masked_vector, 1,
                                                   self.sres.reshape(1,-1) if self.sres.ndim == 1
                                                        else self.sres.reshape(outshape[0], -1)
                                                   ).reshape(outshape)

        # Return the stacked data
        if operation == 'sum':
            return self.wave, self.flux, self.fluxsdev, self.npix, self.ivar, self.sres, self.covar
        return self._get_stack_mean()


        



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Provides a set of file I/O routines.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""

import os
import warnings
import gzip
import shutil
import logging

from IPython import embed

import numpy

from astropy.io import fits
from .log import log_output

def wavelength_vector(npix, header, log10=False):
    """
    Return a vector with wavelength coordinates drawn from the WCS
    coordinates in the header.  The function uses CRVAL1, CRPIX1,
    CDELT1.

    Args:
        npix (:obj:`int`):
            Length of the vector in pixels.
        header (`astropy.io.fits.Header`_):
            Header with the WCS data
        log10 (:obj:`bool`, optional):
            Flag that the wavelengths are logarithmically sampled.

    Returns:
        `numpy.ndarray`_: Wavelengths of a spectrum.

    """
    crval = float(header['CRVAL1'])
    crpix = int(header['CRPIX1'])
    cdelt = float(header['CDELT1'])
    wave = (numpy.arange(1.0,npix+1) - crpix)*cdelt + crval
    return numpy.power(10., wave) if log10 else wave


def readfits_1dspec(filename, log10=False):
    """
    Read a 1D fits spectrum and return two vectors with the wavelength
    and flux.

    Args:
        filename (:obj:`str`):
            Name of the file to read.

    Returns:
        `numpy.ndarray`_: Two numpy.float64 arrays with the
        wavelength and flux read for the spectrum.

    Raises:
        ValueError:
            Raised if the input fits file has more than one extension
            or its primary extension has more than two dimensions.
    """
    hdu = fits.open(filename, mode='readonly')

    if (len(hdu)) != 1:
        warnings.warn('{0} has more than one extension.'.format(filename))
    
    if hdu[0].header['NAXIS'] != 1:
        raise ValueError('{0} is not one dimensional!'.format(filename))
    
    spec = numpy.copy(hdu[0].data).astype(numpy.float64)
    wave = wavelength_vector(spec.size, hdu[0].header, log10=log10)
    hdu.close()
    return wave, spec


def writefits_1dspec(ofile, crval1, cdelt1, flux, hdr=None, overwrite=False):
    """

    Write a simple one-dimensional spectrum.

    Args:
        ofile (:obj:`str`):
            Name of the file to write.
        crval1 (:obj:`float`):
            (Log base 10 of the) Initial wavelength, which is included in the
            header with the keyword 'CRVAL1'; 'CRPIX1' is always set to 1.
        cdelt1 (:obj:`float`):
            The change in (log base 10) wavelength per pixel, which is included
            in the header with the keywords 'CDELT1' and 'CD1_1'; 'CRPIX1' is
            always set to 1.
        flux (`numpy.ndarray`_):
            Vector of the flux values.
        hdr (`astropy.io.fits.Header`_, optional):
            Include sampling data in this header instead of beginning with an
            empty header.
        overwrite (:obj:`bool`, optional):
            Flag to overwrite any existing file of the same name.
    """
    _hdr = fits.Header() if hdr is None else hdr
    _hdr['CRPIX1'] = 1
    _hdr['CRVAL1'] = crval1
    _hdr['CDELT1'] = cdelt1
    _hdr['CD1_1'] = _hdr['CDELT1']
    fits.HDUList([fits.PrimaryHDU(flux, header=_hdr)]).writeto(ofile, overwrite=overwrite)


def read_template_spectrum(filename, data_ext=0, ivar_ext=None, sres_ext=None, log10=False):
    r"""
    Read a template spectrum.
    
    Template spectra are "raw format" files with template data and are,
    at minimum, expected to have the following components::
    
        hdu[0].header['CRVAL1']
        hdu[0].header['CRPIX1']
        hdu[0].header['CDELT1']
        hdu[data_ext].data

    The latter has the flux data. If ``log10`` is true, the
    wavelength solution above is expected to be in log wavelengths.
    
    Args:
        filename (:obj:`str`):
            Name of the fits file to read.
        data_ext (:obj:`str`, optional):
            Name of the extension with the flux data.
        ivar_ext (:obj:`str`, optional):
            Name of the extension with the inverse variance data. If
            None, no inverse data are returned.
        sres_ext (:obj:`str`, optional):
            Name of the extension with the spectral resolution
            (:math:R=\lambda/\delta\lambda`) measurements. If None,
            no spectral resolution data are returned.
        log10 (:obj:`bool`, optional):
            Flag the WCS wavelength coordinates as being in base-10
            log wavelength, instead of linear. Default is to assume
            linear.

    Returns:
        `numpy.ndarray`_: Up to four vectors with the wavelength,
        flux, inverse variance (if `ivar_ext` is provided), and
        spectral resolution (if `sres_ext` is provided) of the
        template spectrum.

    Raises:
        ValueError:
            Raised if fits file is not one-dimensional.
        KeyError:
            Raised if various header keywords or extension names are
            not available.
    """
    with fits.open(filename, mode='readonly') as hdu:
        if len(hdu[data_ext].data.shape) != 1:
            raise ValueError('Spectrum in {0} is not one dimensional!'.format(filename))
    
        spec = numpy.copy(hdu[data_ext].data).astype(numpy.float64)
        wave = wavelength_vector(spec.size, hdu[0].header, log10=log10)

        # Collect objects to return
        ret = (wave, spec)
        if ivar_ext is not None:
            ret += (numpy.copy(hdu[ivar_ext].data).astype(numpy.float64), )
        if sres_ext is not None:
            ret += (numpy.copy(hdu[sres_ext].data).astype(numpy.float64), )
        return ret
    

def init_record_array(shape, dtype):
    r"""
    Utility function that initializes a record array using a provided
    input data type.  For example::

        dtype = [ ('INDX', int, (2,) ),
                  ('VALUE', float) ]

    Defines two columns, one named `INDEX` with two integers per row and
    the one named `VALUE` with a single float element per row.  See
    `numpy.recarray`_.
    
    Args:
        shape (:obj:`int`, :obj:`tuple`):
            Shape of the output array.
        dtype (:obj:`list`):
            List of the tuples that define each element in the record
            array.

    Returns:
        `numpy.recarray`_: Zeroed record array
    """
    data = numpy.zeros(shape, dtype=dtype)
    return data.view(numpy.recarray)


def rec_to_fits_type(rec_element):
    """
    Return the string representation of a fits binary table data type
    based on the provided record array element.
    """
    n = 1 if len(rec_element[0].shape) == 0 else rec_element[0].size
    if rec_element.dtype in [bool, numpy.bool_]:
        return '{0}L'.format(n)
    if rec_element.dtype == numpy.uint8:
        return '{0}B'.format(n)
    if rec_element.dtype == numpy.int16 or rec_element.dtype == numpy.uint16:
        return '{0}I'.format(n)
    if rec_element.dtype == numpy.int32 or rec_element.dtype == numpy.uint32:
        return '{0}J'.format(n)
    if rec_element.dtype == numpy.int64 or rec_element.dtype == numpy.uint64:
        return '{0}K'.format(n)
    if rec_element.dtype == numpy.float32:
        return '{0}E'.format(n)
    if rec_element.dtype == numpy.float64:
        return '{0}D'.format(n)
    
    # If it makes it here, assume its a string
    l = int(rec_element.dtype.str[rec_element.dtype.str.find('U')+1:])
#    return '{0}A'.format(l) if n==1 else '{0}A{1}'.format(l*n,l)
    return '{0}A'.format(l*n)


def rec_to_fits_col_dim(rec_element):
    """
    Return the string representation of the dimensions for the fits
    table column based on the provided record array element.

    The shape is inverted because the first element is supposed to be
    the most rapidly varying; i.e. the shape is supposed to be written
    as row-major, as opposed to the native column-major order in python.
    """
    return None if len(rec_element[0].shape) <= 1 else str(rec_element[0].shape[::-1])


def channel_dictionary(hdu, ext, prefix='C'):
    """
    Construct a dictionary of the channels in a MAPS file.
    """
    channel_dict = {}
    for k, v in hdu[ext].header.items():
        if k[:len(prefix)] == prefix:
            try:
                i = int(k[len(prefix):])-1
            except ValueError:
                continue
            channel_dict[v] = i
    return channel_dict


def channel_units(hdu, ext, prefix='U'):
    """
    Construct an array with the channel units.
    """
    cu = {}
    for k, v in hdu[ext].header.items():
        if k[:len(prefix)] == prefix:
            try:
                i = int(k[len(prefix):])-1
            except ValueError:
                continue
            cu[i] = v.strip()
    channel_units = numpy.empty(max(cu.keys())+1, dtype=object)
    for k, v in cu.items():
        channel_units[k] = v
    return channel_units.astype(str)


def compress_file(ifile, overwrite=False):
    """
    Compress a file using gzip.  The output file has the same name as
    the input file with '.gz' appended.

    Any existing file will be overwritten if overwrite is true.

    An error is raised if the input file name already has '.gz' appended
    to the end.
    """
    if ifile.split('.')[-1] == 'gz':
        raise ValueError('File appears to already have been compressed! {0}'.format(ifile))

    ofile = '{0}.gz'.format(ifile)
    if os.path.isfile(ofile) and not overwrite:
        raise FileExistsError(f'File already exists: {ofile}.\nTo overwrite, set overwrite=True.')

    with open(ifile, 'rb') as f_in:
        with gzip.open(ofile, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)


def create_symlink(ofile, symlink_dir, relative_symlink=True, overwrite=False, loggers=None,
                   quiet=False):
    """
    Create a symlink to the input file in the provided directory.  If
    relative_symlink is True (default), the path to the file is relative
    to the directory with the symlink.
    """
    # Check if the file already exists
    olink_dest = os.path.join(symlink_dir, ofile.split('/')[-1])
    if os.path.isfile(olink_dest) or os.path.islink(olink_dest):
        if overwrite:
            os.remove(olink_dest)
        else:
            return

    # Make sure the symlink directory exists
    if not os.path.isdir(symlink_dir):
        os.makedirs(symlink_dir)

    # Set the relative path for the symlink, if requested
    olink_src = os.path.relpath(ofile, start=os.path.dirname(olink_dest)) \
                    if relative_symlink else ofile
    if not quiet:
        log_output(loggers, 1, logging.INFO, 'Creating symlink: {0}'.format(olink_dest))

    # Create the symlink
    os.symlink(olink_src, olink_dest)




import os
import datetime
import glob
import mmap

from IPython import embed

import numpy

from mangadap.scripts import scriptbase


def has_error(log_file):
    with open(log_file, 'rb', 0) as f, mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ) as s:
        if s.find(b'Traceback') != -1:
            return True
    return False


def sort_errors(err_list):
    tracebacks = []
    for e in err_list:
        with open(e, 'r') as f:
            lines = f.readlines()
        for i,l in enumerate(lines):
            if 'Traceback' in l:
                break
        if i == len(lines):
            continue
        tracebacks += [(lines[0] if i == 1 
                            else (lines[i-2] if 'ERROR' in lines[i-2] else lines[i-1])).strip()]

    unique_traces, inv, counts = numpy.unique(tracebacks, return_inverse=True, return_counts=True)
    return unique_traces, counts, numpy.arange(len(err_list))[inv]
    

def error_report(tracebacks, counts, trace_in_err, err_pltifu, analysis_path, daptypes, ofile):
    with open(ofile, 'w') as f:
        # Print the Tracebacks
        f.write('# Unique Tracebacks\n#\n')
        f.write('# {0:>4} {1:>4} {2}\n'.format('INDX', 'CNT', 'Traceback'))
        for i in range(len(tracebacks)):
            f.write('  {0:>4} {1:>4} {2}\n'.format(i, counts[i], tracebacks[i]))
        f.write('\n\n')

        # Associate each PLATEIFU that failed with its Traceback
        f.write('# Traceback for each PLATEIFU ending in error\n')
        _err_pltifu = numpy.asarray(err_pltifu)
        for i in range(len(tracebacks)):
            f.write('# Traceback {0}: {1}\n'.format(i, tracebacks[i]))
            indx = numpy.asarray(trace_in_err) == i
            for pltifu in _err_pltifu[indx]:
                f.write('{0:>5} {1:>5}\n'.format(*pltifu.split('-')))
            f.write('\n')
        f.write('\n\n')

        # Find if any of the MAPS files were produced for the errored
        # plateifus
        f.write('# MAPS written for errored PLATEIFUs\n')
        s = '# {0:>5} {1:>5}'.format('PLATE', 'IFU')
        for d in daptypes:
            s += ' {0:>30}'.format(d)
        f.write(s + '\n')
        for i in range(len(err_pltifu)):
            plt, ifu = err_pltifu[i].split('-')
            s = '  {0:>5} {1:>5}'.format(plt, ifu)
            for d in daptypes:
                s += ' {0:>30}'.format(os.path.isfile(os.path.join(analysis_path, d, plt, ifu,
                                        'manga-{0}-{1}-MAPS-{2}.fits.gz'.format(plt, ifu, d))))
            f.write(s + '\n')
        f.write('\n\n')


def build_lists(analysis_path, logdir=None):

    if logdir is not None:
        log_path = os.path.join(analysis_path, 'log', logdir)
        print('Searching for status files at:\n    {0}'.format(log_path))

        # Need to ensure the returned lists are sorted and coincident
        ready = glob.glob(os.path.join(log_path, '*', '*', '*.ready'))

        root = ['.'.join(r.split('.')[:-1]) for r in ready]
        started = [r+'.started' if os.path.isfile(r+'.started') else None for r in root]
        done = [r+'.done' if os.path.isfile(r+'.done') else None for r in root]
        err = [r+'.err' if os.path.isfile(r+'.err') else None for r in root]

        return numpy.array(ready, dtype=object), numpy.array(started, dtype=object), \
                numpy.array(done, dtype=object), numpy.array(err, dtype=object)

    # Get and sort the list of directories
    lp = [d for d in os.listdir(os.path.join(analysis_path, 'log')) if 'UTC' in d]
    indx = numpy.argsort(numpy.array([datetime.datetime.strptime(t, '%d%b%YT%H.%M.%SUTC')
                                      for t in lp]))

    # Find all the relevant files
    ndir = len(lp)
    ready = [None]*ndir
    start = [None]*ndir
    done = [None]*ndir
    err = [None]*ndir
    for i,j in enumerate(indx):
        ready[i], start[i], done[i], err[i] = build_lists(analysis_path, logdir=lp[j])

    # Get the list of all possible pltifus
    pltifu = numpy.unique([ '-'.join(os.path.split(f)[1].split('.')[0].split('-')[1:3]) 
                             for f in numpy.concatenate(ready)])
    ncube = len(pltifu)

    # Initialize the lists of the most recent status files
    last_ready = numpy.empty(ncube, dtype=object)
    last_start = numpy.empty(ncube, dtype=object)
    last_done = numpy.empty(ncube, dtype=object)
    last_err = numpy.empty(ncube, dtype=object)

    # Find the status files for the executions that most recently
    # finished
    print('Finding most recently finished executions.')
    for j in range(ncube):
        for i in range(ndir-1,-1,-1):
            indx = numpy.array([e is not None and pltifu[j] in e for e in done[i]])
            if not numpy.any(indx):
                continue
            if numpy.sum(indx) > 1:
                raise ValueError('Degenerate plate ifu')
            k = numpy.where(indx)[0][0]
            last_ready[j] = ready[i][k]
            last_start[j] = start[i][k]
            last_done[j] = done[i][k]
            last_err[j] = err[i][k]
            break

    # For those that did not finish, find the most recent status files
    # for the executions that were only started
    not_finished = numpy.where(last_ready == None)[0]
    print('Finding most recently started executions that did not finish.')
    for j in not_finished:
        for i in range(ndir-1,-1,-1):
            indx = numpy.array([e is not None and pltifu[j] in e for e in start[i]])
            if not numpy.any(indx):
                continue
            if numpy.sum(indx) > 1:
                raise ValueError('Degenerate plate ifu')
            k = numpy.where(indx)[0][0]
            last_ready[j] = ready[i][k]
            last_start[j] = start[i][k]
            last_err[j] = err[i][k]
            break
        
    # For those that were not started, find the most recent status files
    # for the executions that were only ready
    not_started = numpy.where(last_ready == None)[0]
    print('Finding most recent ready executions that did not start.')
    for j in not_started:
        for i in range(ndir-1,-1,-1):
            indx = numpy.array([e is not None and pltifu[j] in e for e in ready[i]])
            if not numpy.any(indx):
                continue
            if numpy.sum(indx) > 1:
                raise ValueError('Degenerate plate ifu')
            k = numpy.where(indx)[0][0]
            last_ready[j] = ready[i][k]
            break

    return last_ready, last_start, last_done, last_err


def dap_status(analysis_path, daptypes, logdir=None):

    ready, started, done, err = build_lists(analysis_path, logdir=logdir)

    # There should be no remaining None values in last_ready
    if numpy.any(ready == None):
        raise ValueError('Should not be any None entries in list of ready files.')

    # Remove any Nones
    done = done[numpy.logical_not(done == None)]
    err = err[numpy.logical_not(err == None)]
    started = started[numpy.logical_not(started == None)]

    log_path = os.path.join(analysis_path, 'log')
    if logdir is not None:
        log_path = os.path.join(log_path, logdir)

    print('Constructing plateifu lists.')
    ready_pltifu = [ '{0}-{1}'.format(*(r.split('/')[-3:-1])) for r in ready ]
    started_pltifu = [ '{0}-{1}'.format(*(r.split('/')[-3:-1])) for r in started ]
    done_pltifu = [ '{0}-{1}'.format(*(r.split('/')[-3:-1])) for r in done ]
    err_pltifu = ['{0}-{1}'.format(*(r.split('/')[-3:-1])) for r in err]

    print('Determine which executions finished in error.')
    errored = [ has_error(e) for e in err ]
    err_pltifu = (numpy.array(err_pltifu)[errored]).tolist() if numpy.any(errored) else []

    print('STATUS:')
    print('      Ready: {0}'.format(len(ready_pltifu)))
    print('    Started: {0}'.format(len(started_pltifu)))
    print('       DONE: {0}'.format(len(done_pltifu)))
    print('    Errored: {0}'.format(len(err_pltifu)))

    tracebacks, counts, trace_in_err = sort_errors(numpy.array(err)[errored])
    print('Errors:')
    for i in range(len(tracebacks)):
        print('{0:>4} {1:>4} {2}'.format(i, counts[i], tracebacks[i]))

    # Print the error report
    ofile = os.path.join(log_path, 'error_report.txt')
    print('Error report written to:\n    {0}'.format(ofile))
    error_report(tracebacks, counts, trace_in_err, err_pltifu, analysis_path, daptypes, ofile)

    print('Comparing lists')
    ready_but_not_started = list(set(ready_pltifu)-set(started_pltifu))
    started_but_not_done = list(set(started_pltifu)-set(done_pltifu))

    print('Ready but not started: {0}'.format(len(ready_but_not_started)))
    print('Started but not DONE: {0}'.format(len(started_but_not_done)))

    if len(ready_but_not_started) > 0:
        ofile = os.path.join(log_path, 'not_started.lst')
        print('PLATEIFUs not started written to:\n    {0}'.format(ofile))
        with open(ofile, 'w') as f:
            for rns in ready_but_not_started:
                f.write('{0:>5} {1:>5}\n'.format(*rns.split('-')))

    if len(started_but_not_done) > 0:
        ofile = os.path.join(log_path, 'not_finished.lst')
        print('PLATEIFUs not completed written to:\n    {0}'.format(ofile))
        with open(ofile, 'w') as f:
            for snd in started_but_not_done:
                f.write('{0:>5} {1:>5}\n'.format(*snd.split('-')))

    if len(err_pltifu) > 0:
        ofile = os.path.join(log_path, 'errored.lst')
        print('PLATEIFUs finished in error written to:\n    {0}'.format(ofile))
        with open(ofile, 'w') as f:
            for err in err_pltifu:
                f.write('{0:>5} {1:>5}\n'.format(*err.split('-')))


class DapStatus(scriptbase.ScriptBase):

    @classmethod
    def name(cls):
        """
        Return the name of the executable.
        """
        return 'dap_status'

    @classmethod
    def get_parser(cls, width=None):

        parser = super().get_parser(description='Check status and log files for DAP progress',
                                    width=width)

        parser.add_argument('plan_file', type=str, help='parameter file with the MaNGA DAP '
                            'execution plan to use instead of the default')

        parser.add_argument('--logdir', type=str, help='Log path (e.g., 31Jan2019T19.16.28UTC)')

        parser.add_argument('--drpver', type=str, help='DRP version', default=None)
        parser.add_argument('--dapver', type=str, help='DAP version', default=None)
        parser.add_argument('--analysis_path', type=str, help='main DAP output path', default=None)

        return parser

    @staticmethod
    def main(args):

        import time
        from mangadap.config import defaults
        from mangadap.par.analysisplan import AnalysisPlanSet
        from mangadap.proc.spatiallybinnedspectra import SpatiallyBinnedSpectra
        from mangadap.proc.stellarcontinuummodel import StellarContinuumModel
        from mangadap.proc.emissionlinemodel import EmissionLineModel

        t = time.perf_counter()

        # Set the the analysis path and make sure it exists
        analysis_path = defaults.dap_analysis_path(drpver=args.drpver, dapver=args.dapver) \
                                if args.analysis_path is None else args.analysis_path

        analysisplan = AnalysisPlanSet.from_par_file(args.plan_file)
        daptypes = []
        for p in analysisplan:
            bin_method = SpatiallyBinnedSpectra.define_method(p['bin_key'])
            sc_method = StellarContinuumModel.define_method(p['continuum_key'])
            el_method = EmissionLineModel.define_method(p['elfit_key'])
            daptypes += [defaults.dap_method(bin_method['key'],
                                            sc_method['fitpar']['template_library_key'],
                                            el_method['continuum_tpl_key'])]

        dap_status(analysis_path, daptypes, logdir=args.logdir)

        print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Container class for the database of emission-line bandpass filters used
for non-parameteric measurements of emission-line flux and velocity
moments.

.. todo::
    Combine this with the main emissionlinemoments.py file.

Class usage examples
--------------------

Emission-line moment databases are defined using SDSS parameter
files. To define a database, you can use one of the default set of
available emission-line moment databases::

    from mangadap.par.emissionmomentsdb import EmissionMomentsDB
    print(EmissionMomentsDB.available_databases())
    elmom = EmissionMomentsDB.from_key('ELBMPL9')

The above call uses the
:func:`~mangadap.par.spectralfeaturedb.SpectralFeatureDB.from_key`
method to define the database using its keyword and the database
provided with the MaNGA DAP source distribution. You can also define
the database directly for an SDSS-style parameter file::

    from mangadap.par.emissionmomentsdb import EmissionMomentsDB
    elmom = EmissionMomentsDB('/path/to/emission/moments/database/myelm.par')

The above will read the file and set the database keyword to 'MYELM'
(i.e., the capitalized root name of the ``*.par`` file). See
:ref:`emissionlines` for the format of the parameter file.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import os
import warnings

import numpy

from pydl.goddard.astro import airtovac
from pydl.pydlutils.yanny import yanny

from .spectralfeaturedb import SpectralFeatureDB
from ..proc.bandpassfilter import BandPassFilterPar
from ..util.datatable import DataTable

#    Table includes:
#
#    .. include:: ../tables/emissionlinemomentsdatatable.rst


class EmissionMomentsDefinitionTable(DataTable):
    """
    Wrapper for an :class:`EmissionMomentsDB`, primarily for output
    to an `astropy.io.fits.BinTableHDU`_.

    .. include:: ../tables/emissionmomentsdefinitiontable.rst
    
    Args:
        name_len (:obj:`int`):
            The maximum length of any of the emission feature names.
        shape (:obj:`int`, :obj:`tuple`, optional):
            The shape of the initial array. If None, the data array
            will not be instantiated; use
            :func:`~mangadap.util.datatable.DataTable.init` to
            initialize the data array after instantiation.
    """
    def __init__(self, name_len=1, shape=None):
        # NOTE: This should require python 3.7 to make sure that this
        # is an "ordered" dictionary.
        datamodel = dict(ID=dict(typ=int, shape=None, descr='Emission feature ID number'),
                         NAME=dict(typ='<U{0:d}'.format(name_len), shape=None,
                                   descr='Name of the emission feature'),
                         RESTWAVE=dict(typ=float, shape=None,
                                       descr='Rest wavelength of the feature'),
                         PASSBAND=dict(typ=float, shape=(2,),
                                       descr='The lower and upper wavelength that bracket the '
                                             'emission feature used for the moment calculations'),
                         BLUEBAND=dict(typ=float, shape=(2,),
                                       descr='The lower and upper wavelength of a passband toward '
                                             'the blue of the emission feature, used to define a '
                                             'linear continuum below the emission feature.'),
                         REDBAND=dict(typ=float, shape=(2,),
                                      descr='The lower and upper wavelength of a passband toward '
                                            'the red of the emission feature, used to define a '
                                            'linear continuum below the emission feature.'))

        keys = list(datamodel.keys())
        super(EmissionMomentsDefinitionTable,
                self).__init__(keys, [datamodel[k]['typ'] for k in keys],
                               element_shapes=[datamodel[k]['shape'] for k in keys],
                               descr=[datamodel[k]['descr'] for k in keys],
                               shape=shape)


class EmissionMomentsDB(SpectralFeatureDB):
    r"""
    Basic container class for the database of emission-line moments
    to calculate.

    Each row of the database is parsed using
    :class:`~mangadap.proc.bandpassfilter.BandPassFilterPar`. For the
    format of the input file, see :ref:`emissionlines-moments`.

    The primary instantiation requires the SDSS parameter file with
    the bandpass data. To instantiate using a keyword (and optionally
    a directory that holds the parameter files), use the
    :func:`~mangadap.par.spectralfeaturedb.SpectralFeatureDB.from_key`
    class method. See the base class for additional attributes.

    Args:
        parfile (:obj:`str`):
            The SDSS parameter file with the database.

    Attributes:
        key (:obj:`str`):
            Database signifying keyword
        file (:obj:`str`):
            File with the data
        size (:obj:`int`):
            Number of features in the database. 
        dummy (`numpy.ndarray`_):
            Boolean array flagging bandpasses as dummy placeholders.
    """
    default_data_dir = 'emission_bandpass_filters'
    def _parse_yanny(self):
        """
        Parse the yanny file (provided by :attr:`file`) for the
        emission-line moment database.

        Returns:
            :obj:`list`: The list of
            :class:`mangadap.par.parset.ParSet` instances for each
            line of the database.
        """
        # Read the yanny file
        par = yanny(filename=self.file, raw=True)
        if len(par['DAPELB']['index']) == 0:
            raise ValueError('Could not find DAPELB entries in {0}!'.format(self.file))

        # Check if any of the bands are dummy bands and warn the user
        self.dummy = numpy.any(numpy.array(par['DAPELB']['blueside']) < 0, axis=1)
        self.dummy |= numpy.any(numpy.array(par['DAPELB']['redside']) < 0, axis=1)
        self.dummy |= numpy.any(numpy.array(par['DAPELB']['primary']) < 0, axis=1)
        if numpy.sum(self.dummy) > 0:
            warnings.warn('Bands with negative wavelengths are used to insert dummy values.'
                          '  Ignoring input bands with indices: {0}'.format(
                                                numpy.array(par['DAPELB']['index'])[self.dummy]))

        # Setup the array of absorption-line index database parameters
        self.size = len(par['DAPELB']['index'])
        parlist = []
        for i in range(self.size):
            invac = par['DAPELB']['waveref'][i] == 'vac'
            parlist += [ BandPassFilterPar(index=par['DAPELB']['index'][i],
                                           name=par['DAPELB']['name'][i],
                                blueside=par['DAPELB']['blueside'][i] if invac \
                                        else airtovac(numpy.array(par['DAPELB']['blueside'][i])),
                                redside=par['DAPELB']['redside'][i] if invac \
                                        else airtovac(numpy.array(par['DAPELB']['redside'][i])),
                                restwave=par['DAPELB']['lambda'][i] if invac \
                                        else airtovac(numpy.array(par['DAPELB']['lambda'][i])),
                                primary=par['DAPELB']['primary'][i] if invac \
                                        else airtovac(numpy.array(par['DAPELB']['primary'][i]))) ]
        return parlist

    def channel_names(self, dicttype=True):
        """
        Return a dictionary with the channel names as the dictionary
        key and the channel number as the dictionary value. If
        ``dicttype`` is False, a list is returned with just the
        string keys.
        """
        channels = [ '{0}-{1}'.format(self.data['name'][i], int(self.data['restwave'][i])) 
                            for i in range(self.size) ]
        return { n:i for i,n in enumerate(channels) } if dicttype else channels

    def to_datatable(self):
        """
        Compile the database with the specifications of each index.
        """
        name_len = 0
        for n in self.data['name']:
            if name_len < len(n):
                name_len = len(n)

        # Instatiate the table data that will be saved defining the set
        # of emission-line moments measured
        db = EmissionMomentsDefinitionTable(name_len=name_len, shape=self.size)
        hk = [ 'ID', 'NAME', 'RESTWAVE', 'PASSBAND', 'BLUEBAND', 'REDBAND' ]
        mk = [ 'index', 'name', 'restwave', 'primary', 'blueside', 'redside' ]
        for _hk, _mk in zip(hk,mk):
            db[_hk] = self.data[_mk]
        return db





# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Container class for the database of absorption-line indices to measure.

Class usage examples
--------------------

Absorption-line index databases are defined using SDSS parameter files. To
define a database, you can use one of the default set of available
absorption-line index databases::

    from mangadap.par.absorptionindexdb import AbsorptionIndexDB
    print(AbsorptionIndexDB.available_databases())
    absdb = AbsorptionIndexDB.from_key('LICKINDX')

The above call uses the
:func:`~mangadap.par.spectralfeaturedb.SpectralFeatureDB.from_key`
method to define the database using its keyword and the database
provided with the MaNGA DAP source distribution. You can also define
the database directly for an SDSS-style parameter file::

    from mangadap.par.absorptionindexdb import AbsorptionIndexDB
    absdb = AbsorptionIndexDB('/path/to/absorption/index/database/myabs.par')

The above will read the file and set the database keyword to
'MYABS' (i.e., the capitalized root name of the ``*.par`` file).
See :ref:`spectralindices` for the format of the parameter file.
    
----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import os
import warnings
import numpy

from pydl.goddard.astro import airtovac
from pydl.pydlutils.yanny import yanny

from .spectralfeaturedb import SpectralFeatureDB
from ..proc.bandpassfilter import BandPassFilterPar

class AbsorptionIndexDB(SpectralFeatureDB):
    r"""
    Basic container class for the database of absorption-line
    indices.

    Each row of the database is parsed using
    :class:`mangadap.proc.bandpassfilter.BandPassFilterPar`.  For the
    format of the input file, see :ref:`spectralindices-absorption`.

    The primary instantiation requires the SDSS parameter file with
    the bandpass data. To instantiate using a keyword (and optionally
    a directory that holds the parameter files), use the
    :func:`~mangadap.par.spectralfeaturedb.SpectralFeatureDB.from_key`
    class method. See the base class for additional attributes.

    Args:
        parfile (:obj:`str`):
            The SDSS parameter file with the database.

    Attributes:
        key (:obj:`str`):
            Database signifying keyword
        file (:obj:`str`):
            File with the data
        size (:obj:`int`):
            Number of features in the database. 
        dummy (`numpy.ndarray`_):
            Boolean array flagging bandpasses as dummy placeholders.
    """
    default_data_dir = 'absorption_indices'
    def _parse_yanny(self):
        """
        Parse the yanny file (provided by :attr:`file`) for the
        bandhead database.

        Returns:
            :obj:`list`: The list of
            :class:`mangadap.par.parset.ParSet` instances for each
            line of the database.
        """
        # Read the yanny file
        par = yanny(filename=self.file, raw=True)
        if len(par['DAPABI']['index']) == 0:
            raise ValueError('Could not find DAPABI entries in {0}!'.format(self.file))

        # Check if any of the bands are dummy bands and warn the user
        self.dummy = numpy.any(numpy.array(par['DAPABI']['blueside']) < 0, axis=1)
        self.dummy |= numpy.any(numpy.array(par['DAPABI']['redside']) < 0, axis=1)
        self.dummy |= numpy.any(numpy.array(par['DAPABI']['primary']) < 0, axis=1)
        if numpy.sum(self.dummy) > 0:
            warnings.warn('Bands with negative wavelengths are used to insert dummy values.'
                          '  Ignoring input bands with indices: {0}'.format(
                                                numpy.array(par['DAPABI']['index'])[self.dummy]))

        # Setup the array of absorption-line index database parameters
        self.size = len(par['DAPABI']['index'])
        parlist = []
        for i in range(self.size):
            invac = par['DAPABI']['waveref'][i] == 'vac'
            comp = par['DAPABI']['component'][i] != 0
            parlist += [ BandPassFilterPar(index=par['DAPABI']['index'][i],
                                           name=par['DAPABI']['name'][i],
                                blueside=par['DAPABI']['blueside'][i] if invac \
                                        else airtovac(numpy.array(par['DAPABI']['blueside'][i])),
                                redside=par['DAPABI']['redside'][i] if invac \
                                        else airtovac(numpy.array(par['DAPABI']['redside'][i])),
                                primary=par['DAPABI']['primary'][i] if invac \
                                        else airtovac(numpy.array(par['DAPABI']['primary'][i])),
                                           units=par['DAPABI']['units'][i],
                                           integrand='flambda',
                                           component=comp) ]
        return parlist

    def channel_names(self, offset=0):
        """
        Return a dictionary with the channel names as the dictionary
        key and the channel number as the dictionary value. An
        ``offset`` can be added to the channel number; i.e., if the
        offset is 2, the channel numbers will be a running number
        starting with 2.
        """
        return {self.data['name'][i] : i + offset for i in range(self.size)}



from IPython import embed

import numpy

from mangadap.util.resolution import match_spectral_resolution
from mangadap.util.constants import DAPConstants

def test_match_spec_res():

    # Build a fake spectrum
    wave = numpy.arange(3650.0, 10501.0, 0.5)

    # Set the resolution to 2.5 angstroms
    fwhm = numpy.zeros(wave.shape, dtype=numpy.float64)
    fwhm += 2.5
    sigma = fwhm/DAPConstants.sig2fwhm
    sres = wave/fwhm

    # Set the flux to a set of uniform emission lines
    flux = numpy.zeros(wave.shape, dtype=numpy.float64)
    gc = numpy.linspace(wave[0], wave[-1], num=20)
    gs = sigma
    for c in gc:
        flux += numpy.exp(-0.5*numpy.square((wave - c)/gs))/numpy.sqrt(2.0*numpy.pi)/gs

    # Set the target spectral resolution
    new_sres = sres/3.0 + 2.0*sres[0]/3.0 + 300
    new_sres = sres-sres+900
    new_fwhm = wave/new_sres
    new_sigma = new_fwhm/DAPConstants.sig2fwhm

    # Set the flux to a set of uniform emission lines
    expected_flux = numpy.zeros(wave.shape, dtype=numpy.float64)
    gc = numpy.linspace(wave[0], wave[-1], num=20)
    for c in gc:
        expected_flux += numpy.exp(-0.5*numpy.square((wave - c)/new_sigma)) \
                                / numpy.sqrt(2.0*numpy.pi)/new_sigma

    # Match the resolution
    new_flux, matched_sres, sigma_offset, new_mask, _ = \
        match_spectral_resolution(wave, flux, sres, wave, new_sres, min_sig_pix=0.0)

#    from matplotlib import pyplot
#    pyplot.plot(wave, flux, label='input')
#    pyplot.plot(wave, new_flux, label='matched resolution')
#    pyplot.plot(wave, expected_flux, label='theoretical result')
#    pyplot.xlabel('Wavelength')
#    pyplot.ylabel('Flux')
#    pyplot.legend()
#    pyplot.show()
#
#    pyplot.plot(wave, flux, label='input')
#    pyplot.plot(wave, new_flux, label='matched resolution')
#    pyplot.plot(wave, expected_flux, label='theoretical result')
#    pyplot.xlim([6875., 6915.])
#    pyplot.xlabel('Wavelength')
#    pyplot.ylabel('Flux')
#    pyplot.legend()
#    pyplot.show()

    assert numpy.mean(new_flux[50:-50]-expected_flux[50:-50]) < 1e-9, 'Failed resolution match!'


